# coding: utf-8
'''
A parser for :title-reference:`GlycoCT{condensed}` format.

:title-reference:`GlycoCT{condensed}` is a multi-line format for representing
glycan structures and compositions published in [1]. The format is intended to
be human-readable, easily compressed, and includes a canonicalization algorithm
to ensure that there is only a single representation for a glycan structure.

:title-reference:`GlycoCT{condensed}` can represent glycan structures with ambiguous
or repeating sub-units. The specification includes additional section directives with
support for stochastic sub-units as well as disjoint subgraphs, though these have not
been implemented in :mod:`glypy`.


References
----------
[1] Herget, S., Ranzinger, R., Maass, K., & Lieth, C.-W. V. D. (2008).
    GlycoCT-a unifying sequence format for carbohydrates.
    Carbohydrate Research, 343(12), 2162–2171.
    https://doi.org/10.1016/j.carres.2008.03.011

'''

import re
import warnings
from collections import defaultdict, Counter, deque, namedtuple, OrderedDict
from functools import cmp_to_key

try:
    from collections.abc import Iterator
except ImportError:
    from collections import Iterator

from glypy.utils import (
    opener, StringIO, root as rootp, tree as treep,
    make_counter, invert_dict, uid,
    RootProtocolNotSupportedError)
from glypy.utils.multimap import OrderedMultiMap
from glypy.structure import monosaccharide, substituent, glycan, Modification, constants, UnknownPosition, NoPosition
from glypy.structure.link import Link, AmbiguousLink
from .format_constants_map import (anomer_map, superclass_map,
                                   link_replacement_composition_map,
                                   modification_map, linkage_type_map)
from .file_utils import ParserError
from .tree_builder_utils import (
    decorate_tree,
    undecorate_tree,
    find_root,
    try_int,
    StructurePrecisionEnum,
    AbstractGraphEntryEnum, NodeCollection)
from glypy.composition import Composition

try:
    range = xrange
except NameError:
    pass


__id = id

Glycan = glycan.Glycan
Monosaccharide = monosaccharide.Monosaccharide
Substituent = substituent.Substituent

Configuration = constants.Configuration
Stem = constants.Stem


START = "!START"
REPINNER = "!REPINNER"
UNDINNER = "!UNDINNER"
RES = "RES"
LIN = "LIN"
REP = "REP"
ALT = "ALT"
UND = "UND"
ISO = "ISO"
NON = "NON"

TERMINAL_STATES = {
    RES,
    LIN,
    ISO,
    NON
}

subsituent_start = "s"
base_start = "b"
repeat_start = "r"
alternative_start = "a"

#: Pattern for parsing the lines of the RES section corresponding
#: to individual |Monosaccharide| residues
res_pattern = re.compile(
    r'''
    (?P<anomer>[abxo])?
    (?P<conf_stem>(?:-[dlx][a-z]+)+)?-?
    (?P<superclass>[A-Z]+)-?
    (?P<indices>[0-9x]+:[0-9x]+)
    (?P<modifications>(\|[0-9x,]+:[0-9a-z]+)+)?
    ''', re.VERBOSE)

#: Pattern for parsing the potentially repeated |Configuration| and |Stem|
#: regions of the lines of the RES section.
conf_stem_pattern = re.compile(r'(?P<config>[dlx])(?P<stem>[a-z]+)')

#: Pattern for parsing modifications found on monosaccharide residue
#: lines in the RES section
modification_pattern = re.compile(r"\|?([0-9,x]+):([^\|;\n]+)")


#: Pattern for parsing |Link| lines found in the LIN section
link_pattern = re.compile(
    r'''(?P<doc_index>\d+)?:
    (?P<parent_residue_index>\d+)
    (?P<parent_atom_replaced>[odhnx])
    \((?P<parent_attachment_position>-?[0-9\-\|]+)[\+\-]
        (?P<child_attachment_position>-?[0-9\-\|]+)\)
    (?P<child_residue_index>\d+)
    (?P<child_atom_replaced>[odhnx])
        ''', re.VERBOSE)


#: Special truncation of the :data:`link_pattern` which is used on
#: REP header sections
internal_link_pattern = re.compile(
    r'''(?P<parent_residue_index>\d+)
    (?P<parent_atom_replaced>[odhnx])
    \((?P<parent_attachment_position>-?[0-9\-\|]+)[\+\-]
        (?P<child_attachment_position>-?[0-9\-\|]+)\)
    (?P<child_residue_index>\d+)
    (?P<child_atom_replaced>[odhnx])
    ''',
    re.VERBOSE)

#: Pattern for interpreting the REP# instance header section
rep_header_pattern = re.compile(
    r'''REP(?P<repeat_index>\d+):
    (?P<internal_linkage>.+)
    =(?P<lower_multitude>-?\d+)-(?P<higher_multitude>-?\d+)''', re.VERBOSE)

repeat_line_pattern = re.compile(r"^(?P<graph_index>\d+)r:r(?P<repeat_index>\d+)")

und_header_pattern = re.compile(r'''UND(?P<und_index>\d+):
    (?P<major>\d+(\.\d*)?):
    (?P<minor>\d+(\.\d*)?)
    ''', re.VERBOSE)

und_link_pattern = re.compile(r'''
    (?P<parent_atom_replaced>[odhnx])
    \((?P<parent_attachment_position>-?[0-9\-\|]+)[\+\-]
        (?P<child_attachment_position>-?[0-9\-\|]+)\)
    (?P<child_atom_replaced>[odhnx])
    ''', re.VERBOSE)


class GlycoCTError(ParserError):
    """Base error for GlycoCT-based parsing exceptions.
    """
    pass


class GlycoCTSectionUnsupported(GlycoCTError):
    '''Indicates that the GlycoCT parser has encountered a section that
    it does not know how to parse.
    '''
    pass


class DeferredRetrieval(object):
    """Callback object to invoke a :class:`GlycoCTGraph` instance's
    :meth:`get_node` method with a set of stored parameters
    at a later time.

    Attributes
    ----------
    direction : AbstractGraphDirectionEnum
        The direction with which to retieve the node
    graph : GlycoCTGraph
        The node graph to call :meth:`GlycoCTGraph.get_node` with
    id : object
        The id of the node to retrieve
    """
    def __init__(self, graph, id, direction=None):
        self.graph = graph
        self.id = id
        self.direction = direction

    def __call__(self):
        return self.graph.get_node(self.id, self.direction)

    def __repr__(self):
        return "{s.__class__.__name__}({s.graph}, {s.id}, {s.direction})".format(
            s=self)


class GlycoCTGraph(object):
    """A graph to store nodes from parsing GlycoCT
    text in.

    Implements a Mapping interface

    Attributes
    ----------
    graph : dict
        Mapping from node id to node-like objects
    id: tuple
        A pair of (class name, ID)
    """
    def __init__(self, graph=None):
        if graph is None:
            graph = dict()
        self.graph = graph
        self.id = (self.__class__, uid())

    def __repr__(self):
        return "{self.__class__.__name__}({self.graph})".format(self=self)

    def __contains__(self, k):
        return k in self.graph

    def __getitem__(self, k):
        return self.get_node(k)

    def __setitem__(self, k, v):
        self.put_node(k, v)

    def __len__(self):
        return len(self.graph)

    def keys(self):
        '''See :meth:`~.Mapping.keys`
        '''
        return self.graph.keys()

    def values(self):
        '''See :meth:`~.Mapping.values`
        '''
        return self.graph.values()

    def items(self):
        '''See :meth:`~.Mapping.items`
        '''
        return self.graph.items()

    def clear(self):
        '''See :meth:`~.MutableMapping.clear`
        '''
        self.graph.clear()

    def __iter__(self):
        return iter(self.graph)

    def get_node(self, id, direction=None):
        """Get a node by its id value.

        Parameters
        ----------
        id : object
            The node's id. Will be case as an :class:`int`
        direction : object, optional
            Included for compatibility, ignored.

        Returns
        -------
        :class:`~.Monosaccharide` or :class:`~.Substituent
        """
        id = int(id)
        return self.graph[id]

    def put_node(self, id, value):
        """Store a node for the given id value.

        Parameters
        ----------
        id : object
            The node's id. Will be case as an :class:`int`
        value : :class:`~.Monosaccharide` or :class:`~.Substituent
            The node to store
        """
        id = int(id)
        self.graph[id] = value

    def form_link(self, parent, child, parent_position, child_position, parent_loss,
                  child_loss, parent_linkage_type=None, child_linkage_type=None, id=None):
        """Form a :class:`~.Link` between `parent` and `child` with the specified parameters.

        If more than one position is passed for `parent_position` or `child_position`, an
        :class:`~.AmbiguousLink` will be created instead.

        Parameters
        ----------
        parent : :class:`~.Monosaccharide` or :class:`~.Substituent
            The parent node in the bond
        child : :class:`~.Monosaccharide` or :class:`~.Substituent
            The child node in the bond
        parent_position : list
            The set of possible positions on the parent node to attach to.
        child_position : list
            The set of possible positions on the child node to attach to.
        parent_loss : str or :class:`~.Compositition`
            The composition lost from the parent node
        child_loss : str or :class:`~.Compositition`
            The composition lost from the child node
        parent_linkage_type : :class:`~.EnumValue`, optional
            A :class:`~.LinkageType` entry describing how the linkage is formed on the parent
        child_linkage_type : :class:`~.EnumValue`, optional
            A :class:`~.LinkageType` entry describing how the linkage is formed on the child
        id : object, optional
            The within-graph unique identifier of the :class:`~.Link` object

        Returns
        -------
        :class:`~.Link`
        """
        if parent.node_type is Substituent.node_type and\
                child.node_type is Monosaccharide.node_type:
            warnings.warn(
                "A monosaccharide with a substituent parent has been detected. "
                "These structures are not fully supported and may not traverse as expected "
                "by default.", stacklevel=7)

        if len(parent_position) > 1 or len(child_position) > 1:
            link_obj = AmbiguousLink(
                parent, child, parent_position=list(map(int, parent_position)),
                child_position=list(map(int, child_position)), parent_loss=parent_loss,
                child_loss=child_loss, id=id, parent_linkage_type=parent_linkage_type,
                child_linkage_type=child_linkage_type)
            link_obj.find_open_position()
        else:
            link_obj = Link(
                parent, child, parent_position=int(parent_position[0]),
                child_position=int(child_position[0]), parent_loss=parent_loss,
                child_loss=child_loss, parent_linkage_type=parent_linkage_type,
                child_linkage_type=child_linkage_type)
        return link_obj

    def deferred_retrieval(self, id, direction=None):
        """Construct a :class:`DeferredRetrieval` instance to carry out the
        :meth:`get_node` at a later time.

        Parameters
        ----------
        id : object
            The node's id. Will be case as an :class:`int`
        direction : object, optional
            Included for compatibility, ignored.

        Returns
        -------
        :class:`DeferredRetrieval`
        """
        return DeferredRetrieval(self, id, direction)

    def __root__(self):
        return self.find_root_nodes()[0]

    def find_root_nodes(self):
        """Find "root" nodes within the graph.

        Returns
        -------
        list
        """
        roots = []
        for _index, node in self.items():
            try:
                if node.parents():
                    continue
            except AttributeError:
                if not isinstance(node, GlycoCTGraph):
                    raise
                else:
                    if rootp(node).parents():
                        continue
            roots.append(node)
        if not roots:
            roots.append(sorted(self.items())[0][1])
        return roots

    def visit(self, node, visited=None, fn=None):
        """Visit `node`, calling `fn` on it, and then call :meth:`visit`
        on each connected node from `node` that had not previously been
        visited (tracked in `visited`).

        If a `node` is actually a :class:`GlycoCTGraph`, it will be
        traversed.

        Parameters
        ----------
        node : :class:`~.Monosaccharide` or :class:`~.Substituent`
            The node to visit.
        visited : :class:`set`, optional
            The set of previously visited node ids. If not provided, an empty set
            will be used.
        fn : :class:`callable`, optional
            The function to call on each node.

        Returns
        -------
        :class:`set`:
            The visited nodes.
        """
        if visited is None:
            visited = set()
        if isinstance(node, GlycoCTGraph):
            for node in node.find_root_nodes():
                self.visit(node, visited, fn)
            return visited

        visited.add(node.id)
        if fn is not None:
            fn(node, visited)
        for _position, link in node.links.items():
            ref = link.to(node)
            if ref.id in visited:
                continue
            else:
                self.visit(ref, visited, fn)
        try:
            for _position, link in node.substituent_links.items():
                ref = link.to(node)
                if ref.id in visited:
                    continue
                else:
                    self.visit(ref, visited, fn)
        except AttributeError:
            pass
        return visited

    def is_fully_connected(self):
        """Check that the graph is fully connected, meaning it has only
        one root node.

        Returns
        -------
        bool
        """
        roots = self.find_root_nodes()
        visited = self.visit(roots[0])
        return len(visited) >= len(self)


class GlycoCTGraphStack(GlycoCTGraph):
    """Represent a stack of :class:`GlycoCTGraph` instances,
    which may be nested inside another graph.

    Attributes
    ----------
    stack: list
        The stack of :class:`GlycoCTGraph` instances in the current state
    history: list
        The historical sequence of :class:`GlycoCTGraph` instances, added to
        but never removed.
    """
    def __init__(self, stack=None, parent=None):  # pylint: disable=super-init-not-called
        if stack is None:
            stack = deque([GlycoCTSubgraph(parent=parent)])
        else:  # pragma: no cover
            _stack = deque()
            _parent = parent
            for level in stack:
                _stack.append(GlycoCTSubgraph(level, parent=_parent))
                _parent = _stack[-1]
            stack = _stack
        self.stack = stack
        self.history = list(stack)
        self.id = (self.__class__, uid())

    @property
    def graph(self):
        """The top :class:`GlycoCTGraph` on the stack

        Returns
        -------
        :class:`GlycoCTGraph`
        """
        return self.stack[-1]

    @property
    def parent(self):
        """The graph that contains this one.

        Returns
        -------
        :class:`GlycoCTGraph`
        """
        return self.stack[0].parent

    @parent.setter
    def parent(self, value):
        self.stack[0].parent = value

    def get_node(self, id, direction=None):
        for level in reversed(self.history):
            try:
                return level.get_node(id, direction)
            except KeyError:
                continue
        raise KeyError(id)

    def deferred_retrieval(self, id, direction=None):
        return DeferredRetrieval(self, id, direction)

    def add(self, subgraph, parent=None):  # pragma: no cover
        """Add `subgraph` to :attr:`stack`, setting the
        graph's parent.

        Parameters
        ----------
        subgraph: :class:`GlycoCTGraph`
            The graph to add to the stack
        parent: :class:`GlycoCTGraph`, optional
            The parent to set for `subgraph`. If :const:`None`, this
            defaults to `self`.

        """
        if subgraph.parent is None:
            if parent is None:
                parent = self
            subgraph.parent = parent
        self.push_level(subgraph)

    def push_level(self, subgraph):
        """Push `subgraph` onto the stack, and the history.

        Parameters
        ----------
        subgraph : :class:`GlycoCTGraph`
            The graph to add.
        """
        self.stack.append(subgraph)
        self.history.append(subgraph)

    def pop_level(self):
        """Pop the last item from :attr:`stack`

        Returns
        -------
        :class:`GlycoCTGraph`
        """
        return self.stack.pop()

    def clear(self):
        self.stack = deque([GlycoCTSubgraph(parent=None)])
        self.history = list(self.stack)

    def find_subgraph_containing(self, id):
        """Find the first subgraph which contains a node with
        the query `id`

        Parameters
        ----------
        id : tuple
            The node id to find.

        Returns
        -------
        :class:`GlycoCTGraph`

        Raises
        ------
        KeyError:
            If the `id` is not found.
        """
        for level in reversed(self.history):
            if id in level:
                return level
        raise KeyError(id)

    def __repr__(self):
        return "{self.__class__.__name__}({self.history})".format(self=self)

    def _count_subgraph_nodes(self):
        layer_relations = defaultdict(list)
        root_layer = self.stack[0]
        for layer in self.stack:
            if layer.parent is None:
                continue
            else:
                key = layer.parent.id
            layer_relations[key].append(layer)

        # repeated subgraphs add to the node count on a layer, but
        # the root layer is also the parent of all other subgraphs
        # which do not have nodes. This is a hacky solution to
        # count just those with nodes in the root layer.
        first_children = layer_relations.pop(root_layer.id, [])
        adjustment = 0
        for child in first_children:
            if isinstance(child, RepeatedGlycoCTSubgraph):
                adjustment += 1
        return sum(map(len, layer_relations.values())) + adjustment

    def __len__(self):
        return sum(map(len, self.stack)) - self._count_subgraph_nodes()

    def find_root_nodes(self):
        return self.stack[0].find_root_nodes()

    def is_fully_connected(self):
        roots = self.find_root_nodes()
        visited = self.visit(roots[0])
        return len(visited) >= len(self)


class GlycoCTSubgraph(GlycoCTGraph):
    """A :class:`GlycoCTGraph` which has an :attr:`parent` for use with
    :class:`GlycoCTGraphStack`.

    Attributes
    ----------
    parent: :class:`GlycoCTGraph`
        The graph that contains this one.

    """
    def __init__(self, graph=None, parent=None):
        super(GlycoCTSubgraph, self).__init__(graph)
        assert not isinstance(parent, dict)
        self.parent = parent

    def postprocess(self):
        """Apply arbitrary post-processing before finalizing the subgraph.

        To be overriden by subclasses. Base implementation is a no-op.
        """
        pass


RepeatedMultitude = namedtuple("RepeatedMultitude", "lower upper")


class RepeatedGlycoCTSubgraph(GlycoCTSubgraph):
    """Implements the machinery for representing a repeated subgraph in GlycoCT.

    Attributes
    ----------
    graph_index: int
    repeast_index: int
        The ``i``th repeating subgraph in the graph.
    internal_linkage: object
        The linkage connecting two repetitions of the subgraph
    external_linkage: object
        The linkage connecting from the final repetition and the
        outside nodes.
    multitude: :class:`RepeatedMultitude`
        Holds the lower and upper range of multiplicities this subgraph may be repeated
        to.
    repetitions: :class:`~.OrderedDict`
        The repetitions of this subgraph, materialized during :meth:`postprocess`
    postponed: :class:`~.deque`
        A queue of post-processing callbacks.

    """
    def __init__(self, graph_index, repeat_index, internal_linkage=None,
                 external_linkage=None, multitude=None, graph=None,
                 parent=None):
        super(RepeatedGlycoCTSubgraph, self).__init__(graph, parent)

        if multitude is None:
            multitude = RepeatedMultitude(-1, -1)
        else:
            multitude = RepeatedMultitude(*multitude)

        self.graph_index = graph_index
        self.repeat_index = repeat_index
        self.internal_linkage = internal_linkage
        self.external_linkage = external_linkage
        self.multitude = multitude
        self.original_graph = None
        self.repetitions = OrderedDict()
        self.postponed = deque()

    def __repr__(self):
        rep = super(RepeatedGlycoCTSubgraph, self).__repr__()
        rep = "%s, %d)" % (rep[:-1], self.repeat_index)
        return rep

    @property
    def terminal_node_index(self):
        """The index of the node that will connect to either the
        external or internal target node.

        Returns
        -------
        int
        """
        return int(self.external_linkage['parent_residue_index'])

    @property
    def last_repeat_index(self):
        """The last repeat's index in :attr:`repetitions`

        Returns
        -------
        int
        """
        sub_unit_indices = sorted(self.repetitions.keys())
        terminal_unit_ix = sub_unit_indices[-1]
        return terminal_unit_ix

    @property
    def terminal_node(self):
        """Retrieves the terminal residue from the subgraph, where
        outgoing connections start from

        Returns
        -------
        MoleculeBase
        """
        terminal_unit_ix = self.last_repeat_index
        parent_graph = self.repetitions[terminal_unit_ix]
        parent = parent_graph.get((terminal_unit_ix, self.terminal_node_index))
        return parent

    @property
    def origin_node_index(self):
        return int(self.external_linkage['child_residue_index'])

    @property
    def first_repeat_index(self):
        """The first repeat's index in :attr:`repetitions`

        Returns
        -------
        int
        """
        sub_unit_indices = sorted(self.repetitions.keys())
        return sub_unit_indices[0]

    @property
    def origin_node(self):
        """Retrieves the root residue from the subgraph, where
        the incoming external connection ends at.

        Returns
        -------
        MoleculeBase
        """
        root_unit_ix = self.first_repeat_index
        child_graph = self.repetitions[root_unit_ix]
        try:
            child = child_graph.get((root_unit_ix, self.origin_node_index))
        except IndexError:
            # the root of this subgraph is nested, coming from an expanded
            # inner subgraph. The index labels recorded in the section header
            # are not accurate anymore. Attempt to recover by returning the
            # root node of this subgraph
            if isinstance(child_graph.root.id[1], tuple):
                return child_graph.root
            else:
                raise

        return child

    def postpone(self, f, args):
        """Queue a callable `f` with `args`
        to be run during postprocessing.

        Parameters
        ----------
        f : :class:`Callable`
            Some callable object
        args : :class:`Iterable`
            The arguments to call `f` with
        """
        self.postponed.append((f, args))

    def complete_postponed_tasks(self):
        """Call all of the postponed tasks.
        """
        i = 0
        while self.postponed:
            i += 1
            f, args = self.postponed.popleft()
            f(*args)

    def __contains__(self, k):
        if self.original_graph is not None:
            return k in self.original_graph
        else:
            return k in self.graph

    @property
    def structure_precision(self):
        if -1 in self.multitude:
            return StructurePrecisionEnum.unknown
        elif self.multitude.lower == self.multitude.upper:
            return StructurePrecisionEnum.exact
        return StructurePrecisionEnum.ranging

    def _build_minigraph(self, graph):
        sub_unit_indices = sorted(map(try_int, graph.keys()))
        for key, node in list(graph.items()):
            if isinstance(node, GlycoCTGraph):
                node.postprocess()
                graph[key] = node.origin_node

        try:
            root = find_root(graph[sub_unit_indices[0]])
        except RootProtocolNotSupportedError:  # pragma: no cover
            raise GlycoCTError("Could not locate subgraph root")
        # glycan_graph = Glycan(root, index_method=None).clone()
        glycan_graph = NodeCollection.from_node(root).clone()
        return glycan_graph

    def _duplicate_graph(self, graph):
        glycan_graph = self._build_minigraph(graph)
        duplicate_graph = {}
        for k, v in graph.items():
            if isinstance(v, GlycoCTSubgraph):
                duplicate_graph[k] = v
            else:
                try:
                    duplicate_graph[k] = glycan_graph.get(v.id)
                except IndexError:
                    # if the subgraph intersperses repeats and
                    # residues, they won't be connected in the
                    # Glycan object, so we must read them back
                    # from the original graph
                    duplicate_graph[k] = graph[k]
        return duplicate_graph

    def postprocess(self, n=None):  # pylint: disable=arguments-differ
        if n is None:
            if self.multitude.upper != -1:
                n = self.multitude.upper
            elif self.multitude.lower != -1:
                n = self.multitude.lower
            else:
                n = 1

        if self.structure_precision is not StructurePrecisionEnum.unknown:
            if not (self.multitude.lower <= n <= self.multitude.upper):  # pragma: no cover
                raise GlycoCTError("{} is not within the range of {}".format(n, self.multitude))

        self.original_graph = self._duplicate_graph(self.graph)
        glycan_graph = self._build_minigraph(self.graph)

        graph = OrderedDict({1: glycan_graph.clone(index_method=None)})
        decorate_tree(graph[1], 1)

        parent_residue_index = self.terminal_node_index
        parent_atom_replaced = link_replacement_composition_map[self.internal_linkage["parent_atom_replaced"]]
        parent_linkage_type = linkage_type_map[self.internal_linkage["parent_atom_replaced"]]
        parent_attachment_position = self.internal_linkage["parent_attachment_position"]

        child_residue_index = self.origin_node_index
        child_atom_replaced = link_replacement_composition_map[self.internal_linkage["child_atom_replaced"]]
        child_linkage_type = linkage_type_map[self.internal_linkage["child_atom_replaced"]]
        child_attachment_position = self.internal_linkage["child_attachment_position"]

        op_stack = []

        for i in range(2, n + 1):
            graph[i] = glycan_graph.clone(index_method=None)
            graph[i] = decorate_tree(graph[i], i)
            parent_graph = graph[i - 1]
            child_graph = graph[i]
            parent_node = parent_graph.get((i - 1, parent_residue_index))
            child_node = child_graph.get((i, child_residue_index))
            op_stack.append(
                (self.form_link, [parent_node, child_node],
                 dict(parent_position=parent_attachment_position,
                      child_position=child_attachment_position,
                      parent_loss=parent_atom_replaced,
                      child_loss=child_atom_replaced,
                      parent_linkage_type=parent_linkage_type,
                      child_linkage_type=child_linkage_type)))

        for op in op_stack:
            f, args, kwargs = op
            f(*args, **kwargs)

        self.repetitions = graph
        self.complete_postponed_tasks()

    def handle_incoming_link(self, parent_getter, parent_position, parent_loss,
                             child_position, child_loss, id=None):
        child = self.origin_node
        parent = parent_getter()
        if parent_loss == Composition("H"):
            child_loss = Composition("OH")

        self.form_link(
            parent, child, parent_position=parent_position, child_position=child_position,
            parent_loss=parent_loss, child_loss=child_loss, id=id)

    def handle_outgoing_link(self, child_getter, parent_position, parent_loss,
                             child_position, child_loss, id=None):
        child = child_getter()
        parent = self.terminal_node
        if isinstance(child, RepeatedGlycoCTSubgraph):
            child.handle_incoming_link(
                lambda: parent, parent_position=parent_position,
                child_position=child_position, parent_loss=parent_loss, child_loss=child_loss, id=id)
        else:
            self.form_link(
                parent, child, parent_position=parent_position, child_position=child_position,
                parent_loss=parent_loss, child_loss=child_loss, id=id)

    def handle_abstract_subgraph_link(self, parent_getter, child_getter, parent_position,
                                      parent_loss, child_position, child_loss, id=None):
        parent = parent_getter()
        child = child_getter()
        self.form_link(
            parent, child, parent_position=parent_position,
            child_position=child_position, parent_loss=parent_loss,
            child_loss=child_loss, id=id)

    def get_node(self, id, direction=None):
        id = int(id)
        if direction == AbstractGraphEntryEnum.parent:
            child_graph = self.repetitions[self.first_repeat_index]
            child = child_graph.get((self.first_repeat_index, id))
            return child
        elif direction == AbstractGraphEntryEnum.child:
            parent_graph = self.repetitions[self.last_repeat_index]
            parent = parent_graph.get((self.last_repeat_index, id))
            return parent
        elif direction == AbstractGraphEntryEnum.internal or direction is None:
            return self.graph[id]
        else:
            raise Exception("Unknown direction %s" % direction)

    def __root__(self):  # pragma: no cover
        root_node = find_root(self.repetitions[self.first_repeat_index])
        return root_node

    def find_root_nodes(self):
        if not self.repetitions:
            return super(RepeatedGlycoCTSubgraph, self).find_root_nodes()
        roots = []
        for _rep_index, node_set in self.repetitions.items():
            for node in node_set:
                try:
                    if node.parents():
                        continue
                except AttributeError:
                    if not isinstance(node, GlycoCTGraph):
                        raise
                    else:
                        if rootp(node).parents():
                            continue
                roots.append(node)
        if not roots:
            roots.append(sorted(self.items())[0][1])
        return roots

    def prepare_glycan(self):
        subglycan_start = self.repetitions[self.first_repeat_index]
        subglycan_start.deindex()
        return Glycan(subglycan_start.root, index_method=None)


UndeterminedProbability = namedtuple("UndeterminedProbability", "major minor")


LinkageSpecification = namedtuple(
    "LinkageSpecification", [
        "id", "parent_residue_index", "parent_atom_replaced", "parent_attachment_position",
        "child_residue_index", "child_atom_replaced", "child_attachment_position",
        "parent_linkage_type", "child_linkage_type"
    ])


class UndeterminedGlycoCTSubgraph(GlycoCTSubgraph):
    def __init__(self, und_index, probability=None, parent_ids=None,
                 subtree_linkages=None, graph=None, parent=None):
        super(UndeterminedGlycoCTSubgraph, self).__init__(graph, parent)
        if probability is None:
            probability = UndeterminedProbability(100., 100.)
        if parent_ids is None:
            parent_ids = []
        if subtree_linkages is None:
            subtree_linkages = []
        self.parent_ids = parent_ids
        self.subtree_linkages = subtree_linkages

    def __root__(self):
        return self[sorted(self.keys())[0]]

    def postprocess(self):
        parents = [
            self.parent.get_node(parent_id) for parent_id in self.parent_ids]
        try:
            child = [rootp(self)]
        except RootProtocolNotSupportedError:  # pragma: no cover
            raise GlycoCTError("Could not locate subgraph root")

        linkage = self.subtree_linkages[0]
        parent_loss = linkage["parent_atom_replaced"]
        parent_position = linkage["parent_attachment_position"]
        child_loss = linkage["child_atom_replaced"]
        child_position = linkage["child_attachment_position"]

        link_obj = AmbiguousLink(
            parents, child, parent_position=list(map(int, parent_position)),
            child_position=list(map(int, child_position)), parent_loss=parent_loss,
            child_loss=child_loss)
        try:
            link_obj.find_open_position()
        except ValueError:  # pragma: no cover
            if link_obj.child_position == 1 and Modification.Acidic in link_obj.child.modifications[1]:
                link_obj.child_position = 2
                ix = link_obj.child_position_choices.index(1)
                link_obj.child_position_choices.pop(ix)
                link_obj.child_position_choices.insert(ix, 2)
                link_obj.apply()
                link_obj.find_open_position()
            else:
                raise


def _build_graph(glycoct_str):  # pragma: no cover
    rep = StringIO(glycoct_str)
    inst = GlycoCTReader(rep, completes=False)
    return next(inst)


def extract_composition(parser):
    from glypy.structure.glycan_composition import (
        GlycanComposition, MonosaccharideResidue, SubstituentResidue)
    store = GlycanComposition()
    # remove links between layers in the stack
    for layer in list(parser.stack)[1:]:
        node = rootp(layer)
        for _position, link in list(node.links.items()):
            if link.is_child(node):
                link.break_link(refund=True)

    for layer in parser.stack:
        for node in layer.find_root_nodes():
            if isinstance(node, Monosaccharide):
                node = MonosaccharideResidue.from_monosaccharide(node)
            elif isinstance(node, Substituent):
                node = SubstituentResidue(node.name)
            else:
                raise ValueError(node)
            store[node] += 1
    return store


class GlycoCTReader(GlycoCTGraphStack, Iterator):
    """Parse :title-reference:`GlycoCT{condensed}` text data into |Glycan| objects.

    The parser implements the :class:`Iterator` interface, yielding successive glycans
    from a text stream separated by empty lines.

    The parser can understand fully specified and partially ambiguous structures.
    When :attr:`allow_repeats` is |True| and a ``REP`` section is encountered, it
    will be expanded to its minimum multiplicity, or 1 if the minimum is unknown.
    ``UND`` sections will be connected to the main graph by :class:`~.AmbiguousLink`
    instead of :class:`~.Link` objects.

    Attributes
    ----------
    allow_repeats : :class:`bool`
        Whether or not to permit ``REP`` sections. Defaults to |True|
    completes : :class:`bool`
        Whether or not to translate the built graph into a |Glycan| object. Defaults
        to |True|
    handle : file-like
        The text file being read from
    in_repeat : :class:`bool`
        Indicates the parser is currently parsing a ``REP`` section's sub-graph
    in_undetermined : bool
        Indicates the parser is currently parsing a ``UND`` section's sub-graph
    postponed : list
        Holds all the deferred operations for the top-most graph as :class:`callable`
        objects
    root : :class:`Monosaccharide`
        The root node of the produced graph
    state : str
        The current state of the parser's state machine
    structure_class : type
        The |Glycan| sub-class to produce
    repeats : dict
        Maps RES section index to :class:`RepeatedGlycoCTSubgraph`
    undetermineds : dict
        Maps UND section index to :class:`UndeterminedGlycoCTSubgraph`
    """

    @classmethod
    def loads(cls, glycoct_str, structure_class=Glycan, allow_repeats=True):
        '''Parse results from |str|'''
        rep = StringIO(glycoct_str)
        return cls(rep, structure_class=structure_class, allow_repeats=allow_repeats)

    def __init__(self, stream, structure_class=Glycan, allow_repeats=True, completes=True):
        super(GlycoCTReader, self).__init__()

        self._state = None
        self.state = START

        self.completes = completes

        self.handle = opener(stream, "r")
        self.in_repeat = False
        self.in_undetermined = False
        self.repeats = {}
        self.undetermineds = {}
        self.postponed = []

        self.root = None
        self._iter = None

        self.allow_repeats = allow_repeats
        self.structure_class = structure_class

        self._index = 0
        self._source_line = 0
        self._segment_iterator = None

        self._output_queue = deque()

    def _read(self):
        for line in self.handle:
            self._source_line += 1
            for segment in re.split(r"\s|;", line):
                if not segment.strip():
                    continue
                self._current_segment = segment
                yield segment

    def _reset(self):
        self.clear()
        self.root = None
        self.postponed = []
        self.repeats.clear()
        self.undetermineds.clear()
        self.in_repeat = False
        self._index += 1

    def reset(self):
        if self.completes:
            self._reset()

    def __iter__(self):
        '''
        Calls :meth:`parse` and stores it for reuse with :meth:`__next__`
        '''
        if self._iter is None:
            self._iter = self.parse()
        return self._iter

    def next(self):
        '''
        Calls :meth:`parse` if the internal iterator has not been instantiated
        '''
        if self._iter is None:
            iter(self)
        return next(self._iter)

    #: Alias for next. Supports Py3 Iterator interface
    __next__ = next

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    def _parse_modifications(self, residue_dict):
        mods = residue_dict["modifications"]
        modifications = OrderedMultiMap()
        if mods is not None:
            for modp, mod in modification_pattern.findall(mods):
                positions = modp.split(",")
                if len(positions) > 1:
                    warnings.warn("Multi-site Modifications are not fully supported")
                for p in positions:
                    modifications[try_int(p)] = modification_map[mod]
        is_reduced = "aldi" in modifications[1]
        if is_reduced:
            modifications.pop(1, "aldi")
            is_reduced = monosaccharide.ReducedEnd()
        else:
            is_reduced = None
        return modifications, is_reduced

    def _parse_conf_stem(self, residue_dict):
        conf_stem = residue_dict["conf_stem"]
        if conf_stem is not None:
            config, stem = zip(*conf_stem_pattern.findall(conf_stem))
        else:
            config = ('x',)
            stem = ('x',)
        stem_ = tuple(Stem[s] for s in stem)
        configuration_ = tuple(Configuration[c] for c in config)
        return stem_, configuration_

    def handle_residue_line(self, line):
        '''
        Handle a base line, creates an instance of |Monosaccharide|
        and adds it to :attr:`graph` at the given index.

        Called by :meth:`parse`
        '''
        _, ix, residue_str = re.split(r"^(\d+)b", line, maxsplit=1)
        residue_dict = res_pattern.search(residue_str).groupdict()

        modifications, is_reduced = self._parse_modifications(residue_dict)

        stem_, configuration_ = self._parse_conf_stem(residue_dict)

        ring_start_, ring_end_ = [
            (try_int(i) if i != 'x' else UnknownPosition) for i in residue_dict["indices"].split(":")]

        anomer_ = anomer_map[residue_dict['anomer']]
        super_class_ = superclass_map[residue_dict['superclass']]
        ix = int(ix)
        residue = monosaccharide.Monosaccharide(
            fast=True, stem=stem_, modifications=modifications,
            reduced=is_reduced, configuration=configuration_,
            ring_start=ring_start_, ring_end=ring_end_, anomer=anomer_,
            superclass=super_class_, id=ix)

        self.put_node(ix, residue)
        if self.root is None:
            self.root = residue

    def handle_residue_substituent(self, line):
        '''
        Handle a substituent line, creates an instance of |Substituent|
        and adds it to :attr:`graph` at the given index. The |Substituent| object is not yet linked
        to a |Monosaccharide| instance.

        Called by :meth:`parse`

        '''
        _, ix, subsituent_str = re.split(r"^(\d+)s:", line, maxsplit=1)
        sub = Substituent(subsituent_str.strip())

        self[int(ix)] = sub

    def handle_blank(self):
        self._complete_structure()

    def enter_res(self):
        if self.state in (START, REPINNER, UNDINNER):
            pass
        elif self.state in TERMINAL_STATES:
            self.in_repeat = False
            self._complete_structure()
        else:
            raise GlycoCTError("Invalid State Transition at line %d" % self._source_line)

        self.state = RES

    def enter_lin(self):
        if self.state != RES:
            raise GlycoCTError("LIN before RES at line %d" % self._source_line)
        self.state = LIN

    def enter_rep(self):
        if not self.allow_repeats:
            raise GlycoCTSectionUnsupported(
                "Repeat are not allowed (set allow_repeats=True to allow them) at line %d" % self._source_line)
        self.state = REP
        self.in_repeat = True

    def enter_und(self):
        self.state = UND
        self.in_undetermined = True

    def parse_link(self, line):
        link_dict = link_pattern.search(line)
        if link_dict is not None:
            link_dict = link_dict.groupdict()
        else:
            raise GlycoCTError("Could not interpret link", line)
        id = link_dict['doc_index']
        parent_residue_index = int(link_dict['parent_residue_index'])
        child_residue_index = int(link_dict['child_residue_index'])

        parent_atom_replaced = link_replacement_composition_map[link_dict["parent_atom_replaced"]]
        parent_attachment_position = list(map(int, link_dict["parent_attachment_position"].split("|")))
        try:
            parent_linkage_type = linkage_type_map[link_dict['parent_atom_replaced']]
        except KeyError:
            parent_linkage_type = constants.LinkageType.x

        child_atom_replaced = link_replacement_composition_map[link_dict["child_atom_replaced"]]
        child_attachment_position = list(map(int, link_dict["child_attachment_position"].split("|")))
        try:
            child_linkage_type = linkage_type_map[link_dict['child_atom_replaced']]
        except KeyError:
            child_linkage_type = constants.LinkageType.x

        return LinkageSpecification(
            id, parent_residue_index, parent_atom_replaced, parent_attachment_position,
            child_residue_index, child_atom_replaced, child_attachment_position,
            parent_linkage_type, child_linkage_type)

    def handle_linkage(self, line):
        '''
        Handle a linkage line, creates an instance of |Link| and
        attaches it to the two referenced nodes in :attr:`graph`. The parent node is always
        an instance of |Monosaccharide|, and the child node
        may either be an instance of |Monosaccharide| or
        |Substituent| or |Monosaccharide|.

        Called by :meth:`parse`

        See also |Link| for more information on the impact of instantiating
        a |Link| object.
        '''
        id, parent_residue_index, parent_atom_replaced, parent_attachment_position,\
            child_residue_index, child_atom_replaced, child_attachment_position,\
            parent_linkage_type, child_linkage_type = self.parse_link(line)

        parent = self.get_node(parent_residue_index)
        child = self.get_node(child_residue_index)

        is_parent_repeat = isinstance(parent, RepeatedGlycoCTSubgraph)
        is_child_repeat = isinstance(child, RepeatedGlycoCTSubgraph)

        if is_parent_repeat and is_child_repeat:
            inner = max([parent, child], key=lambda x: x.repeat_index)
            if child == inner:

                def child_getter():
                    return child.origin_node

                def parent_getter():
                    return parent.get_node(
                        parent.terminal_node_index,
                        AbstractGraphEntryEnum.internal)
            else:

                def child_getter():
                    return child.get_node(
                        child.origin_node_index,
                        AbstractGraphEntryEnum.internal)

                def parent_getter():
                    return parent.terminal_node
            if parent_atom_replaced == child_atom_replaced == Composition('H'):
                parent_atom_replaced = Composition('H')
                child_atom_replaced = Composition('OH')
            inner.postpone(
                inner.handle_abstract_subgraph_link,
                (parent_getter,
                 child_getter,
                 parent_attachment_position, parent_atom_replaced,
                 child_attachment_position, child_atom_replaced, id)
            )
        elif is_parent_repeat:
            parent.postpone(parent.handle_outgoing_link, (
                self.deferred_retrieval(child_residue_index),
                parent_attachment_position, parent_atom_replaced,
                child_attachment_position, child_atom_replaced, id)
            )
        elif is_child_repeat:
            child.postpone(child.handle_incoming_link, (
                self.deferred_retrieval(parent_residue_index),
                parent_attachment_position, parent_atom_replaced,
                child_attachment_position, child_atom_replaced, id)
            )
        else:
            self.form_link(
                parent, child,
                parent_position=parent_attachment_position, child_position=child_attachment_position,
                parent_loss=parent_atom_replaced, child_loss=child_atom_replaced, id=id,
                parent_linkage_type=parent_linkage_type, child_linkage_type=child_linkage_type)

    def handle_repeat_stub(self, line):
        if not self.allow_repeats:
            raise GlycoCTSectionUnsupported(
                "Repeat are not allowed (set allow_repeats=True to allow them)")

        match = repeat_line_pattern.search(line).groupdict()

        graph_index = try_int(match['graph_index'])
        repeat_index = try_int(match["repeat_index"])

        repeat = RepeatedGlycoCTSubgraph(
            int(graph_index), int(repeat_index), parent=self.graph)
        repeat._index = self._index

        self[graph_index] = repeat
        self.repeats[repeat_index] = repeat

        if self.root is None:
            self.root = repeat

    def handle_repeat_inner(self, line):
        if not self.in_repeat:
            raise GlycoCTError(
                "Encountered %r outside of REP at line %d" % (
                    line, self._source_line))
        header_dict = rep_header_pattern.search(line).groupdict()

        repeat_index = int(header_dict['repeat_index'])
        repeat_record = self.repeats[repeat_index]

        self.push_level(repeat_record)

        linkage = internal_link_pattern.search(header_dict['internal_linkage']).groupdict()
        repeat_record.internal_linkage = linkage
        repeat_record.external_linkage = linkage
        repeat_record.multitude = RepeatedMultitude(
            try_int(header_dict['lower_multitude']),
            try_int(header_dict['higher_multitude']))
        self.state = REPINNER

    def handle_und_inner(self, line):
        if not self.in_undetermined:
            raise GlycoCTError("Encountered %r outside of UND at line %d" % (
                line, self._source_line))
        header_dict = und_header_pattern.search(line).groupdict()
        parent_line = next(self._segment_iterator)
        subtree_linkage_line = next(self._segment_iterator)

        ids = list(map(int, parent_line.split(":")[1].split("|")))
        subtree_linkages = []
        match = und_link_pattern.search(subtree_linkage_line.split(":")[1])
        if match is None:
            raise GlycoCTError("Could not interpret UND SubtreeLinkage %r at line %d" % (
                subtree_linkage_line, self._source_line))
        else:
            link_dict = match.groupdict()
            link_dict["parent_atom_replaced"] = link_replacement_composition_map[
                link_dict["parent_atom_replaced"]]
            link_dict["parent_attachment_position"] = list(
                map(int, link_dict["parent_attachment_position"].split("|")))
            link_dict["child_atom_replaced"] = link_replacement_composition_map[
                link_dict["child_atom_replaced"]]
            link_dict["child_attachment_position"] = list(
                map(int, link_dict["child_attachment_position"].split("|")))
            subtree_linkages.append(link_dict)

        und_index = int(header_dict['und_index'])
        prob = UndeterminedProbability(float(header_dict['major']), float(header_dict['minor']))
        record = UndeterminedGlycoCTSubgraph(
            und_index, prob, parent_ids=ids,
            subtree_linkages=subtree_linkages, parent=self)
        self.undetermineds[und_index] = record
        self.push_level(record)
        self.state = UNDINNER

    def _complete_structure(self):
        if self.completes:
            result = self.postprocess()
            if result is not None:
                self._output_queue.append(result)
            self.reset()
        else:
            self._output_queue.append(self)

    def postprocess(self):
        '''
        Handle all deferred operations such as binding together and expanding
        repeating units. Removes any distinguishing markers on node ids, and
        constructs a new instance of :attr:`structure_class` from the accumulated
        graph

        Returns
        -------
        Glycan
        '''
        for level in reversed(self.history):
            level.postprocess()

        for postop in self.postponed:
            postop[0](*postop[1:])

        if self.root is None:
            return None

        if self.is_fully_connected():
            try:
                inst = undecorate_tree(
                    self.structure_class(
                        root=rootp(self.root), index_method=None)
                ).reindex()
                return inst
            except RootProtocolNotSupportedError:  # pragma: no cover
                raise GlycoCTError("Could not locate graph root")
        else:
            # warnings.warn("The parsed structure was not fully connected. Producing a Composition")
            return extract_composition(self)

    def parse(self):
        '''
        Returns an iterator that yields each complete :class:`Glycan` instance
        from the underlying text stream.
        '''

        # Create a reference to the segment iterator
        # as late as possible, but bind it to the object
        # state so it can be referenced independent of this
        # outermost loop
        self._segment_iterator = self._read()

        for line in self._segment_iterator:
            if line.strip() == "":
                self.handle_blank()
                while self._output_queue:
                    yield self._output_queue.popleft()

            elif RES == line.strip():
                self.enter_res()
                while self._output_queue:
                    yield self._output_queue.popleft()

            elif LIN == line.strip():
                self.enter_lin()

            elif REP == line.strip():
                self.enter_rep()

            elif UND == line.strip():
                self.enter_und()

            # REP definition block
            elif line.strip()[:3] == REP:
                self.handle_repeat_inner(line)
            elif line.strip()[:3] == UND:
                self.handle_und_inner(line)
            elif ALT == line.strip():
                raise GlycoCTSectionUnsupported(ALT)

            elif re.search(r"^(\d+)b", line) and self.state == RES:
                self.handle_residue_line(line)
            elif re.search(r"^(\d+)s:", line) and self.state == RES:
                self.handle_residue_substituent(line)
            elif re.search(r"^(\d+)r:", line) and self.state == RES:
                self.handle_repeat_stub(line)
            elif re.search(r"^(\d+):(\d+)", line) and self.state == LIN:
                self.handle_linkage(line)
            else:
                raise GlycoCTError("Unknown format error: %s on line %d" % (line, self._source_line))

        if self.root is not None:
            self._complete_structure()
            while self._output_queue:
                yield self._output_queue.popleft()


GlycoCT = GlycoCTReader


def read(stream, structure_class=Glycan, allow_repeats=True):
    '''
    A convenience wrapper for :class:`GlycoCTReader`
    '''
    return GlycoCTReader(stream, structure_class=structure_class, allow_repeats=allow_repeats)


def load(stream, structure_class=Glycan, allow_repeats=True, allow_multiple=True):  # pragma: no cover
    """Read all structures from the provided text stream.

    Parameters
    ----------
    stream : file-like
        The text stream to parse structures from
    structure_class : type, optional
        :class:`~.Glycan` subclass to use
    allow_repeats : bool, optional
        Whether or not to allow ``REP`` sections

    Returns
    -------
    :class:`~.Glycan` or :class:`list` of :class:`~.Glycan`
    """
    g = GlycoCTReader(stream, structure_class=structure_class, allow_repeats=allow_repeats)
    first = next(g)
    if not allow_multiple:
        return first
    second = None
    try:
        second = next(g)
        collection = [first, second]
        collection.extend(g)
        return collection
    except StopIteration:
        return first


def loads(text, structure_class=Glycan, allow_repeats=True, allow_multiple=True):
    """Read all structures from the provided text string.

    Parameters
    ----------
    text : str
        The text to parse structures from
    structure_class : type, optional
        :class:`~.Glycan` subclass to use
    allow_repeats : bool, optional
        Whether or not to allow ``REP`` sections

    Returns
    -------
    :class:`~.Glycan` or :class:`list` of :class:`~.Glycan`
    """

    text_buffer = StringIO(text)
    return load(text_buffer, structure_class, allow_repeats, allow_multiple)


def detect_glycoct(string):
    return string.lstrip()[:3] == "RES"


invert_anomer_map = invert_dict(anomer_map)
invert_superclass_map = invert_dict(superclass_map)


class DictTree(object):
    def __init__(self, state=START, store=None):
        if store is None:
            store = {}
        self.store = defaultdict(dict, store)
        self.state = state

    def __getitem__(self, key):
        for subtree in self.store.values():
            try:
                return subtree[key]
            except KeyError:
                continue
        return self.store[key]

    def __setitem__(self, key, value):
        self.store[self.state][key] = value

    def __len__(self):
        return sum(map(len, self.store))

    def __iter__(self):
        return iter(self.store)

    def keys(self):
        return self.store.keys()

    def items(self):
        vals = list(self.store.values())
        acc = vals[0].items()
        for v in vals[1:]:
            acc = acc + v.items()
        return acc

    def __contains__(self, key):
        return key in self.store

    def get(self, key, subtree=None, default=None):
        if subtree is None:
            subtree = self.state
        return self.store[subtree].get(key, default)


class GlycoCTWriterBase(object):
    """Summary

    Attributes
    ----------
    buffer : file-like
        The buffer to write structures to. If :attr:`nobuffer` is |True|,
        this will be a :class:`~.StringIO` object which will be returned
        on each write.
    dependencies : :class:`defaultdict` of :class:`dict`
        Track the relationships between child nodes and their parents.
        Used during linkage writing
    full : :class:`bool`
        Whether or not to traverse :class:`~.Monosaccharide`-:class:`~.Monosaccharide`
        linkages.
    index_to_residue : :class:`DictTree`
        A state-specific mapping from index to :attr:`~.Monosaccharide.id`.
    lin_accumulator : list
        Accumulator list of :class:`~.Link` objects.
    lin_counter : function
        A stateful counter which when called returns the next
        integer in a sequence used to index entries in the `LIN` section
    nobuffer : bool
        Whether or not the writer was initialized with a write-able buffer
    res_counter : function
        A stateful counter which when called returns the next
        integer in a sequence used to index entries in the `RES` section
    residue_to_index : :class:`DictTree`
        A state-specific mapping from :attr:`~.Monosaccharide.id` to index.
    state : str
        The current state of the writer
    structure : :class:`~.SaccharideCollection`
        The structure currently being written. May be a :class:`~.Monosaccharide`,
        :class:`~.Glycan`, :class:`~.GlycanComposition`.
    und_counter : function
        A stateful counter which when called returns the next
        integer in a sequence used to index `UND` sections.
    """

    def __init__(self, structure=None, buffer=None, full=True):
        self.nobuffer = False
        if buffer is None:
            self.nobuffer = True
            buffer = StringIO()

        self.buffer = buffer
        self.structure = structure
        self.full = full

        self.state = START
        self._initialize_counters()
        self._initialize_index_tree()
        self._initialize_link_trackers()

    def _initialize_counters(self):
        self.res_counter = make_counter()
        self.lin_counter = make_counter()
        self.und_counter = make_counter()

    def _initialize_index_tree(self):
        # Look-ups for mapping RES nodes to objects by section index and id,
        # respectively
        self.index_to_residue = DictTree(self.state)
        self.residue_to_index = DictTree(self.state)

    def _initialize_link_trackers(self):
        # Accumulator for linkage indices and mapping linkage indices to
        # dependent RES indices
        self.lin_accumulator = []
        self.dependencies = defaultdict(dict)

    @property
    def structure(self):
        return self._structure

    @structure.setter
    def structure(self, value):
        if value is None:
            self._structure = value
            return
        try:
            structure = treep(value)
        except TypeError:
            try:
                root = rootp(value)
                structure = Glycan(root, index_method=None)
            except TypeError:
                raise TypeError("Could not extract or construct a tree structure from %r" % value)
        self._structure = structure

    def _reset(self):
        self.state = START
        self._initialize_counters()
        self._initialize_index_tree()
        self._initialize_link_trackers()
        if self.nobuffer:
            self.buffer = StringIO()

    def _glycoct_sigils(self, link):
        '''
        Helper method for determining which GlycoCT symbols and losses to present
        '''
        parent_loss_str, child_loss_str = link._glycoct_sigils()

        return parent_loss_str, child_loss_str

    def handle_link(self, link, ix, parent_ix, child_ix):
        parent_loss_str, child_loss_str = self._glycoct_sigils(link)

        if link.has_ambiguous_linkage():
            rep = "{ix}:{parent_ix}{parent_loss}({parent_position}+{child_position}){child_ix}{child_loss}"
            return rep.format(
                ix=ix,
                parent_ix=parent_ix,
                parent_loss=parent_loss_str,
                parent_position='|'.join(map(str, link.parent_position_choices)),
                child_ix=child_ix,
                child_loss=child_loss_str,
                child_position='|'.join(map(str, link.child_position_choices)))
        else:
            rep = "{ix}:{parent_ix}{parent_loss}({parent_position}+{child_position}){child_ix}{child_loss}"
            return rep.format(
                ix=ix,
                parent_ix=parent_ix,
                parent_loss=parent_loss_str,
                parent_position=link.parent_position,
                child_ix=child_ix,
                child_loss=child_loss_str,
                child_position=link.child_position)

    def handle_substituent(self, substituent):  # pylint: disable=redefined-outer-name
        return "s:{0}".format(substituent.name.replace("_", "-"))

    def _format_monosaccharide(self, monosaccharide):  # pylint: disable=redefined-outer-name
        residue_template = "{ix}b:{anomer}{conf_stem}{superclass}-{ring_start}:{ring_end}{modifications}"

        # This index is reused many times
        monosaccharide_index = self.res_counter()

        # Format individual fields
        anomer = invert_anomer_map[monosaccharide.anomer]
        conf_stem = ''.join("-{0}{1}".format(c.name, s.name)
                            for c, s in zip(monosaccharide.configuration, monosaccharide.stem))
        if None in monosaccharide.configuration and None in monosaccharide.stem:
            conf_stem = ''
        superclass = "-" + invert_superclass_map[monosaccharide.superclass]

        modifications = '|'.join(
            "{0}:{1}".format(k, v.name) for k, v in monosaccharide.modifications.items())
        null_positions = (UnknownPosition, NoPosition)
        modifications = "|" + modifications if modifications != "" else ""
        ring_start = monosaccharide.ring_start if monosaccharide.ring_start not in null_positions else 'x'
        ring_end = monosaccharide.ring_end if monosaccharide.ring_end not in null_positions else 'x'

        # The complete monosaccharide residue line
        residue_str = residue_template.format(ix=monosaccharide_index, anomer=anomer, conf_stem=conf_stem,
                                              superclass=superclass, modifications=modifications,
                                              ring_start=ring_start, ring_end=ring_end)
        return residue_str, monosaccharide_index

    def handle_monosaccharide(self, monosaccharide):  # pylint: disable=redefined-outer-name
        residue_str, monosaccharide_index = self._format_monosaccharide(monosaccharide)
        res = [residue_str]
        lin = []
        visited_subst = dict()
        # Construct the substituent lines
        # and their links
        for _lin_pos, link_obj in monosaccharide.substituent_links.items():
            sub = link_obj.to(monosaccharide)
            if sub.id not in visited_subst:
                sub_index = self.res_counter()
                subst_str = str(sub_index) + self.handle_substituent(sub)
                res.append(subst_str)
                visited_subst[sub.id] = sub_index
            lin.append(
                self.handle_link(
                    link_obj, self.lin_counter(), monosaccharide_index, visited_subst[sub.id]))

        return [res, lin, monosaccharide_index]

    def handle_glycan(self, structure):  # pragma: no cover
        if structure is None:
            raise GlycoCTError("No structure is ready to be written.")

        self.lin_accumulator = []
        self.buffer.write("RES\n")

        visited = set()
        for node in (structure):
            if node.id in visited:
                continue
            visited.add(node.id)
            res, lin, index = self.handle_monosaccharide(node)

            self.lin_accumulator.append((index, lin))
            self.residue_to_index[node.id] = index
            self.index_to_residue[index] = node

            if self.full:
                for _pos, lin in node.links.items():
                    if lin.is_child(node):
                        continue
                    self.dependencies[lin.child.id][node.id] = ((self.lin_counter(), lin))
            for line in res:
                self.buffer.write(line + '\n')

            # If this serialization is not meant to be full
            # do not visit residues beyond the first.
            if not self.full:
                break

        self.buffer.write("LIN\n")
        for res_ix, links in self.lin_accumulator:
            for line in links:
                self.buffer.write(line + '\n')
            residue = self.index_to_residue[res_ix]
            if self.full:
                for _pos, lin in residue.links.items():
                    if lin.is_child(residue):
                        continue
                    child_res = lin.child
                    ix, lin = self.dependencies[child_res.id][residue.id]
                    self.buffer.write(
                        self.handle_link(lin, ix, res_ix, self.residue_to_index[child_res.id]) + "\n")
        return self.buffer

    def begin_underdetermined(self):
        if self.state != UND:
            self.buffer.write("UND\n")
            self.state = UND
            self.index_to_residue.state = UND
            self.residue_to_index.state = UND

    def _format_subtree_linkage(self, linkage_args):
        (parent_link_type, parent_position,
         child_position, child_link_type) = linkage_args
        return "%s(%d+%d)%s" % (parent_link_type, parent_position,
                                child_position, child_link_type)

    def _get_viable_und_parents(self):
        valid_parent_inds = []
        for index, node in self.index_to_residue[START].items():
            if node.node_type == Monosaccharide.node_type:
                valid_parent_inds.append(index)
        return valid_parent_inds

    def handle_und_header(self, major_probability=100.0, minor_probability=100.0, parent_ids=None,
                          subtree_linkage_args=None):
        if parent_ids is None:
            parent_ids = list(self._get_viable_und_parents())

        index = self.und_counter()
        self.buffer.write("UND%d:%0.1f:%0.1f\n" % (index, major_probability, minor_probability))
        self.buffer.write(
            "ParentIDs:%s\n" % ('|'.join(map(str, parent_ids))))
        self.buffer.write("SubtreeLinkageID1:%s\n" % (
            self._format_subtree_linkage(subtree_linkage_args)))

    def dump(self):
        buffer = self.handle_glycan(self.structure)
        if self.nobuffer:
            value = buffer.getvalue()
            self._reset()
            return value
        return buffer

    def write(self, structure):
        self.structure = structure
        self._reset()
        return self.dump()

    def embed(self, writer):
        writer.res_counter = self.res_counter
        writer.lin_counter = self.lin_counter
        writer.und_counter = self.und_counter
        writer.index_to_residue = self.index_to_residue
        writer.buffer = self.buffer
        writer.state = self.state
        return writer

    def _determine_und_linkage_type_glycan_composition(self, glycan_composition):
        if len(glycan_composition) == 1:
            keys = list(glycan_composition)
            key = keys[0]
            if key.node_type == Substituent.node_type:
                return "d", "n"
            else:
                return "o", "d"
        else:
            return "o", "d"

    def add_glycan_composition(self, glycan_composition):
        for m in OrderingComparisonContext(self).sort_residues(glycan_composition):
            for _i in range(glycan_composition[m]):
                self.add_glycan_composition_single({m: 1})

    def add_glycan_composition_single(self, glycan_composition):
        self.begin_underdetermined()
        linkage_types = self._determine_und_linkage_type_glycan_composition(glycan_composition)
        self.handle_und_header(subtree_linkage_args=(linkage_types[0], -1, -1, linkage_types[1]))
        writer = GlycanCompositionGlycoCTWriter(
            glycan_composition, self.buffer)
        self.embed(writer)
        writer.handle_glycan(glycan_composition)


class GlycanCompositionGlycoCTWriter(GlycoCTWriterBase):
    def __init__(self, structure=None, buffer=None, full=True, standardize=False):
        super(GlycanCompositionGlycoCTWriter, self).__init__(structure, buffer, full)
        self.standardize = standardize

    @property
    def structure(self):
        return self._structure

    def _standardize_substituent_linkage(self, link):
        if self.standardize:
            from glypy.io.nomenclature import identity
            from glypy.structure.named_structures import monosaccharides
            try:
                if link.child.name == 'n_acetyl':
                    if identity.is_a(link.parent, monosaccharides.HexNAc):
                        if link.parent_position == -1:
                            link.parent_position = 2
            except AttributeError:
                pass

    @structure.setter
    def structure(self, value):
        from glypy.structure.glycan_composition import GlycanComposition

        if value is None:
            self._structure = value
            return
        if isinstance(value, (GlycanComposition, dict)):
            value = GlycanComposition(value)
        else:
            try:
                structure = treep(value)
            except TypeError:
                try:
                    root = rootp(value)
                    structure = Glycan(root, index_method=None)
                except TypeError:
                    raise TypeError("Could not extract or construct a tree structure from %r" % value)
            value = GlycanComposition.from_glycan(structure)
        self._structure = value

    def _unspool(self, mapping):
        sorter = OrderingComparisonContext(self)
        order = sorter.sort_residues(mapping.keys(), reverse=True)
        for key in order:
            count = mapping[key]
            if count < 1:
                continue
            for _i in range(count):
                yield key

    def _write_und_subgraph(self, substituent):  # pylint: disable=redefined-outer-name
        self.handle_und_header(subtree_linkage_args=('o', -1, 1, 'n'))
        self.buffer.write("RES\n")
        sub_index = self.res_counter()
        subst_str = str(sub_index) + self.handle_substituent(substituent)
        self.buffer.write("%s\n" % subst_str)

    def handle_link(self, link, ix, parent_ix, child_ix):
        self._standardize_substituent_linkage(link)
        return super(GlycanCompositionGlycoCTWriter, self).handle_link(link, ix, parent_ix, child_ix)

    def handle_glycan(self, structure):
        if structure is None:
            raise GlycoCTError("No structure is ready to be written.")

        from glypy.structure.glycan_composition import (
            SubstituentResidue, MolecularComposition)

        self.lin_accumulator = []

        disconnected_substituents = []
        molecules = []

        nodes = []

        for node in self._unspool(structure):
            if isinstance(node, SubstituentResidue):
                disconnected_substituents.append(node)
                continue
            elif isinstance(node, MolecularComposition):
                molecules.append(node)
                continue
            else:
                nodes.append(node)

        if nodes:
            self.buffer.write("RES\n")
            for node in nodes:
                res, lin, index = self.handle_monosaccharide(node)

                self.lin_accumulator.append((index, lin))
                self.residue_to_index[node.id] = index
                self.index_to_residue[index] = node

                if self.full:
                    for _pos, lin in node.links.items():
                        if lin.is_child(node):
                            continue
                        self.dependencies[lin.child.id][node.id] = ((self.lin_counter(), lin))
                for line in res:
                    self.buffer.write(line + '\n')

                # If this serialization is not meant to be full
                # do not visit residues beyond the first.
                if not self.full:
                    break
            # if self.lin_accumulator:
            if any(linkages for res_ix, linkages in self.lin_accumulator):
                self.buffer.write("LIN\n")
                for res_ix, links in self.lin_accumulator:
                    for line in links:
                        self.buffer.write(line + '\n')
                    residue = self.index_to_residue[res_ix]
                    if self.full:
                        for _pos, lin in residue.links.items():
                            if lin.is_child(residue):
                                continue
                            child_res = lin.child
                            ix, lin = self.dependencies[child_res.id][residue.id]
                            self.buffer.write(
                                self.handle_link(lin, ix, res_ix, self.residue_to_index[child_res.id]) + "\n")
            if disconnected_substituents:
                self.begin_underdetermined()
                for node in disconnected_substituents:
                    self._write_und_subgraph(node)
        elif disconnected_substituents:
            for subst in disconnected_substituents:
                self.buffer.write("RES\n")
                sub_index = self.res_counter()
                subst_str = str(sub_index) + self.handle_substituent(subst)
                self.buffer.write("%s\n" % subst_str)

        if molecules:
            raise TypeError("Cannot serialize MolecularComposition to GlycoCT")

        return self.buffer


def all_node_depth(node, visited=None):
    if visited is None:
        visited = set()
    if node.id in visited:  # pragma: no cover
        return 0
    visited.add(node.id)
    depth_count = 1
    children = list(node.children())
    try:
        children += list(node.substituents())
    except AttributeError:
        pass
    if children:
        depth_count += max(all_node_depth(ch, visited) for p, ch in children)
    return depth_count


class OrderingComparisonContext(object):
    def __init__(self, parent):
        self.parent = parent
        if isinstance(self.structure, Glycan) and not self.structure.has_index():
            self.structure.reindex()
        self.branch_to_terminal_count = self.build_branch_to_terminal_count()

    @property
    def structure(self):
        return self.parent.structure

    def get_branch_from_link_label(self, link):
        return link.label[0]

    def build_branch_to_terminal_count(self):
        counter = Counter()
        try:
            for key in sorted(self.structure.branch_parent_map.keys(), reverse=True):
                parent = self.structure.branch_parent_map[key]
                counter[parent] += counter[key] + 1
        except AttributeError:
            pass
        return counter

    def _residue_diff(self, res_a, res_b):
        n_child_residues_a = all_node_depth(res_a)
        n_child_residues_b = all_node_depth(res_b)
        diff_child_res = n_child_residues_a - n_child_residues_b

        if diff_child_res != 0:
            if diff_child_res < 0:
                diff_child_res = -1
            else:
                diff_child_res = 1

        try:
            branch_length_a = max((all_node_depth(cr) for p, cr in res_a.children()))
        except ValueError:
            branch_length_a = 0
        try:
            branch_length_b = max((all_node_depth(cr) for p, cr in res_b.children()))
        except ValueError:
            branch_length_b = 0

        diff_longest_branch = branch_length_a - branch_length_b

        if diff_longest_branch != 0:
            if diff_longest_branch < 0:
                diff_longest_branch = -1
            else:
                diff_longest_branch = 1

        n_branches_from_a = 0
        n_branches_from_b = 0
        for link in res_a.links.values():
            if link.is_parent(res_a):
                branch_label = self.get_branch_from_link_label(link)
                n_branches_from_a = max(n_branches_from_a, self.branch_to_terminal_count[branch_label])

        for link in res_b.links.values():
            if link.is_parent(res_b):
                branch_label = self.get_branch_from_link_label(link)
                n_branches_from_b = max(n_branches_from_b, self.branch_to_terminal_count[branch_label])
        diff_n_branches_from = n_branches_from_a - n_branches_from_b

        if diff_n_branches_from != 0:
            if diff_n_branches_from < 0:
                diff_n_branches_from = -1
            else:
                diff_n_branches_from = 1

        if res_a == res_b:
            subtree_diff = 0
        else:
            subtree_a = GlycoCTWriter(Glycan.subtree_from(self.structure, res_a)).dump()
            subtree_b = GlycoCTWriter(Glycan.subtree_from(self.structure, res_b)).dump()
            subtree_diff = (subtree_b > subtree_a) - (subtree_b < subtree_a)
            # cmp(subtree_b, subtree_a)
        return (diff_child_res, diff_longest_branch, diff_n_branches_from, subtree_diff, subtree_a, subtree_b)

    def _compare_residue_ordering(self, res_a, res_b):
        n_child_residues_a = all_node_depth(res_a)
        n_child_residues_b = all_node_depth(res_b)
        diff_child_res = n_child_residues_a - n_child_residues_b

        if diff_child_res != 0:
            if diff_child_res < 0:
                return -1
            else:
                return 1

        try:
            branch_length_a = max((all_node_depth(cr) for p, cr in res_a.children()))
        except ValueError:
            branch_length_a = 0
        try:
            branch_length_b = max((all_node_depth(cr) for p, cr in res_b.children()))
        except ValueError:
            branch_length_b = 0

        diff_longest_branch = branch_length_a - branch_length_b

        if diff_longest_branch != 0:
            if diff_longest_branch < 0:
                return -1
            else:
                return 1

        n_branches_from_a = 0
        n_branches_from_b = 0
        for link in res_a.links.values():
            if link.is_parent(res_a):
                branch_label = self.get_branch_from_link_label(link)
                n_branches_from_a = max(n_branches_from_a, self.branch_to_terminal_count[branch_label])

        for link in res_b.links.values():
            if link.is_parent(res_b):
                branch_label = self.get_branch_from_link_label(link)
                n_branches_from_b = max(n_branches_from_b, self.branch_to_terminal_count[branch_label])
        diff_n_branches_from = n_branches_from_a - n_branches_from_b

        if diff_n_branches_from != 0:
            if diff_n_branches_from < 0:
                return -1
            else:
                return 1

        if res_a == res_b:
            return 0

        subtree_a = GlycoCTWriter(Glycan.subtree_from(self.structure, res_a)).dump()
        subtree_b = GlycoCTWriter(Glycan.subtree_from(self.structure, res_b)).dump()
        return (subtree_b > subtree_a) - (subtree_b < subtree_a)

    def compare_residue_ordering(self, res_a, res_b):
        ordered = self._compare_residue_ordering(res_a, res_b)
        return ordered

    def _link_diff(self, link_a, link_b):  # pragma: no cover
        parent_pos_a = link_a.parent_position
        parent_pos_b = link_b.parent_position
        try:
            diff_parent = parent_pos_a - parent_pos_b
        except TypeError as e:
            print(parent_pos_a, parent_pos_b, link_a, link_b)
            raise e

        if diff_parent != 0:
            if diff_parent < 0:
                diff_parent = -1
            else:
                diff_parent = 1

        child_pos_a = link_a.child_position
        child_pos_b = link_b.child_position
        diff_child = child_pos_a - child_pos_b

        if diff_child != 0:
            if diff_child < 0:
                diff_child = -1
            else:
                diff_child = 1

        sigils_a = link_a._glycoct_sigils()
        sigils_b = link_b._glycoct_sigils()

        diff_sig0 = 0
        if sigils_a[0] != sigils_b[0]:
            diff_sig0 = ord(sigils_a[0]) - ord(sigils_b[0])
            if diff_sig0 < 0:
                diff_sig0 = -1
            else:
                diff_sig0 = 1

        diff_sig1 = 0
        if sigils_a[1] != sigils_b[1]:
            diff_sig1 = ord(sigils_a[1]) - ord(sigils_b[1])
            if diff_sig1 < 0:
                diff_sig1 = -1
            else:
                diff_sig1 = 1

        child_a = link_a.child
        child_b = link_b.child
        ordered = self.compare_residue_ordering(child_a, child_b)
        return (diff_parent, diff_child, diff_sig0, diff_sig1, ordered)

    def _compare_link_ordering(self, link_a, link_b):
        # Ignoring # of links for now since it is difficult
        # to compute
        parent_pos_a = link_a.parent_position
        parent_pos_b = link_b.parent_position
        try:
            diff_parent = parent_pos_a - parent_pos_b
        except TypeError as e:
            print(parent_pos_a, parent_pos_b, link_a, link_b)
            raise e

        if diff_parent != 0:
            if diff_parent < 0:
                return -1
            else:
                return 1

        child_pos_a = link_a.child_position
        child_pos_b = link_b.child_position
        diff_child = child_pos_a - child_pos_b

        if diff_child != 0:
            if diff_child < 0:
                return -1
            else:
                return 1

        sigils_a = link_a._glycoct_sigils()
        sigils_b = link_b._glycoct_sigils()

        if sigils_a[0] != sigils_b[0]:
            diff_sig0 = ord(sigils_a[0]) - ord(sigils_b[0])
            if diff_sig0 < 0:
                return -1
            else:
                return 1

        if sigils_a[1] != sigils_b[1]:
            diff_sig1 = ord(sigils_a[1]) - ord(sigils_b[1])
            if diff_sig1 < 0:
                return -1
            else:
                return 1

        child_a = link_a.child
        child_b = link_b.child
        ordered = self.compare_residue_ordering(child_a, child_b)
        return ordered

    def compare_link_ordering(self, link_a, link_b):
        ordered = self._compare_link_ordering(link_a, link_b)
        return ordered

    def sort_links(self, links, reverse=False):
        return sorted(links, key=cmp_to_key(self.compare_link_ordering),
                      reverse=reverse)

    def sort_residues(self, residues, reverse=False):
        return sorted(residues, key=cmp_to_key(self.compare_residue_ordering),
                      reverse=reverse)


class SubtreeJourney(object):
    def __init__(self, links):
        self.links = links

    def __repr__(self):
        template = "{self.__class__.__name__}({self.links!r})"
        return template.format(self=self)

    def __iter__(self):
        return iter(self.links)

    def __len__(self):
        return len(self.links)

    def __getitem__(self, i):
        return self.links[i]


class OrderedSubtreeTraverser(object):
    def __init__(self, ordering_context):
        self.ordering_context = ordering_context

    def visit_link(self, link, ignore_ambiguous=False):
        if isinstance(link, AmbiguousLink) and len(link.parent_choices) and not ignore_ambiguous:
            subtree = OrderedSubtreeTraverser(self.ordering_context)
            return [SubtreeJourney(subtree.visit_subtree(link, True))]
        node = link.child
        if node.node_type is Monosaccharide.node_type:
            next_links = self.visit_monosaccharide(node)
        else:
            next_links = self.visit_substituent(node)
        return next_links

    def visit_monosaccharide(self, monosaccharide):
        link_collection = list(monosaccharide.substituent_links.values())
        link_collection.extend(
            [cl for p, cl in monosaccharide.children(links=True)])
        links = self.ordering_context.sort_links(link_collection)
        return links[::-1]

    def visit_substituent(self, substituent):
        links = self.ordering_context.sort_links(
            [cl for p, cl in substituent.children(links=True)])
        return links[::-1]

    def visit_subtree(self, link, ignore_ambiguous=False):
        visited = set()
        link_queue = deque([link])
        journey = deque()

        while link_queue:
            link = link_queue.popleft()
            # Explicitly add before skipping to avoid double-writing
            # residues, but including multiple-link cases
            journey.append(link)
            if isinstance(link, SubtreeJourney):
                continue
            if link.child.id in visited:
                continue
            visited.add(link.child.id)
            link_queue.extend(
                self.visit_link(
                    link, ignore_ambiguous=ignore_ambiguous))
            ignore_ambiguous &= False
        return journey


class OrderRespectingGlycoCTWriter(GlycoCTWriterBase):
    def __init__(self, structure, buffer=None, full=True):
        super(OrderRespectingGlycoCTWriter, self).__init__(structure, buffer, full)

        self.ordering_context = OrderingComparisonContext(self)
        self.link_queue = deque()

    def handle_monosaccharide(self, monosaccharide):
        residue_str, monosaccharide_index = self._format_monosaccharide(monosaccharide)

        self.index_to_residue[monosaccharide_index] = monosaccharide
        self.residue_to_index[monosaccharide.id] = monosaccharide_index

        link_collection = list(monosaccharide.substituent_links.values())
        if self.full:
            link_collection.extend([cl for p, cl in monosaccharide.children(links=True)])

        links = self.ordering_context.sort_links(link_collection)
        self.link_queue.extendleft(links[::-1])
        return residue_str

    def handle_substituent(self, substituent):
        substituent_index = self.res_counter()

        self.index_to_residue[substituent_index] = substituent
        self.residue_to_index[substituent.id] = substituent_index

        subst_str = "%ss:%s" % (substituent_index, substituent.name.replace("_", "-"))

        links = self.ordering_context.sort_links([cl for p, cl in substituent.children(links=True)])
        self.link_queue.extendleft(links[::-1])
        return subst_str

    def handle_glycan(self, structure):
        if structure is None:
            raise GlycoCTError("No structure is ready to be written.")

        self.process_graph(structure.root)
        return self.buffer

    def process_graph(self, root, visited=None, link_queue=None):
        if visited is None:
            visited = set()

        if link_queue is None:
            link_queue = self.link_queue

        links_in_order = []
        underdetermined_subtrees = []

        self.buffer.write("RES\n")
        if root.node_type is Monosaccharide.node_type:
            res_str = self.handle_monosaccharide(root)
            self.buffer.write(res_str + "\n")
        else:
            res_str = self.handle_substituent(root)
            self.buffer.write(res_str + "\n")

        while link_queue:
            link = link_queue.popleft()
            # Explicitly add before skipping to avoid double-writing
            # residues, but including multiple-link cases
            links_in_order.append(link)
            if link.child.id in visited:
                continue
            visited.add(link.child.id)
            if link.child.node_type is Monosaccharide.node_type:
                line = self.handle_monosaccharide(link.child)
            else:
                line = self.handle_substituent(link.child)
            self.buffer.write(line + "\n")

        self.buffer.write("LIN\n")
        for link in links_in_order:
            if not link.is_substituent_link() and not self.full:
                continue
            parent_ix = self.residue_to_index[link.parent.id]
            child_ix = self.residue_to_index[link.child.id]
            line = self.handle_link(
                link, self.lin_counter(), parent_ix, child_ix)
            self.buffer.write(line + "\n")

        return visited, underdetermined_subtrees


class UNDOrderRespectingGlycoCTWriter(OrderRespectingGlycoCTWriter):
    def gobble_subtree(self, link):
        traverser = OrderedSubtreeTraverser(self.ordering_context)
        return SubtreeJourney(traverser.visit_subtree(link, True))

    def process_graph(self, root, visited=None, link_queue=None):
        if visited is None:
            visited = set()

        if link_queue is None:
            link_queue = self.link_queue

        links_in_order = []
        underdetermined_subtrees = []

        self.buffer.write("RES\n")
        if root.node_type is Monosaccharide.node_type:
            res_str = self.handle_monosaccharide(root)
            self.buffer.write(res_str + "\n")
        else:
            res_str = self.handle_substituent(root)
            self.buffer.write(res_str + "\n")

        while link_queue:
            link = link_queue.popleft()
            if isinstance(link, AmbiguousLink) and len(link.parent_choices) > 1:
                underdetermined_subtrees.append(self.gobble_subtree(link))
                continue
            # Explicitly add before skipping to avoid double-writing
            # residues, but including multiple-link cases
            links_in_order.append(link)
            if link.child.id in visited:
                continue
            visited.add(link.child.id)
            if link.child.node_type is Monosaccharide.node_type:
                line = self.handle_monosaccharide(link.child)
            else:
                line = self.handle_substituent(link.child)
            self.buffer.write(line + "\n")

        self.buffer.write("LIN\n")
        for link in links_in_order:
            if not link.is_substituent_link() and not self.full:
                continue
            parent_ix = self.residue_to_index[link.parent.id]
            child_ix = self.residue_to_index[link.child.id]
            line = self.handle_link(
                link, self.lin_counter(), parent_ix, child_ix)
            self.buffer.write(line + "\n")

        return visited, underdetermined_subtrees

    def handle_glycan(self, structure):
        if structure is None:
            raise GlycoCTError("No structure is ready to be written.")

        visited, underdetermined_subtrees = self.process_graph(structure.root)

        if underdetermined_subtrees:
            self.begin_underdetermined()
            for i, und in enumerate(underdetermined_subtrees, 1):
                und = deque(und)
                self.buffer.write("UND%d:100.0:100.0\n" % i)
                base_link = und.popleft()
                parent_ids = "|".join([str(self.residue_to_index[p.id]) for p in base_link.parent_choices])
                self.buffer.write("ParentIDs:%s\n" % parent_ids)
                subtree_linkage = self.handle_link(base_link, '-', '', '')[2:]
                self.buffer.write("SubtreeLinkageID1:%s\n" % subtree_linkage)

                outer_link_deque = self.link_queue
                self.link_queue = deque([])

                child = base_link.child
                self.process_graph(child, visited=visited, link_queue=und)

                self.link_queue = outer_link_deque

        return self.buffer


GlycoCTWriter = UNDOrderRespectingGlycoCTWriter


def dump(structure, buffer=None):
    '''
    Serialize the |Glycan| into :title-reference:`GlycoCT{condensed}`, using
    `buffer` to store the result. If `buffer` is |None|, then the
    function will operate on a newly created :class:`StringIO` object.

    Parameters
    ----------
    structure: |Glycan|
        The structure to serialize
    buffer: file-like or None
        The stream to write the serialized structure to. If |None|, uses an instance
        of :class:`StringIO`

    Returns
    -------
    file-like or str if ``buffer`` is :const:`None`
    '''
    from glypy import GlycanComposition
    if isinstance(structure, GlycanComposition):
        return GlycanCompositionGlycoCTWriter(structure, buffer).dump()
    return GlycoCTWriter(structure, buffer).dump()


def dumps(structure):
    '''
    Serialize the |Glycan| into :title-reference:`GlycoCT{condensed}`, returning
    the text as a string.

    Parameters
    ----------
    structure: |Glycan|
        The structure to serialize

    Returns
    -------
    str
    '''
    from glypy import GlycanComposition
    if isinstance(structure, GlycanComposition):
        return GlycanCompositionGlycoCTWriter(structure, None).dump()
    return GlycoCTWriter(structure, None).dump()


def _postprocessed_single_monosaccharide(monosaccharide, convert=True):
    if convert:
        monostring = GlycoCTWriterBase(monosaccharide, None, full=False).dump()
    else:
        monostring = monosaccharide
    monostring = monostring.replace("\n", " ")
    if monostring.endswith("LIN "):
        monostring = monostring.replace(" LIN ", "")
    else:
        monostring = monostring.strip()
    return monostring


Monosaccharide.register_serializer("glycoct", _postprocessed_single_monosaccharide)
Glycan.register_serializer("glycoct", dumps)


def canonicalize(structure):
    return loads(dumps(structure))



from collections import namedtuple

import numpy as np
import matplotlib

from matplotlib.path import Path
from matplotlib.textpath import TextPath
import matplotlib.patches as patches
from matplotlib.transforms import Affine2D

from .common import MonosaccharidePatch
from .geometry import centroid, bounding_box

from glypy.utils.enum import Enum


default_line_color = 'black'
default_scale_factor = 1.0
zorder = 2
line_weight = 0.5


NodeLayoutTransforms = namedtuple("NodeLayoutTransforms", (
    'layout_transform', 'substituent_transform', 'annotation_transform'))


class UnknownShapeException(Exception):
    pass


def draw_circle(ax, x, y, color, scale=0.1):
    path = Path(Path.unit_circle().vertices * scale, Path.unit_circle().codes)
    trans = matplotlib.transforms.Affine2D().translate(x, y)
    t_path = path.transformed(trans)
    patch = patches.PathPatch(
        t_path, facecolor=color.value, lw=line_weight, zorder=2)
    a = ax.add_patch(patch)
    ma = MonosaccharidePatch(saccharide_shape=(a,))
    return ma


def draw_square(ax, x, y, color, scale=0.1):
    square_verts = np.array([
        (0.5, 0.5),
        (0.5, -0.5),
        (-0.5, -0.5),
        (-0.5, 0.5),
        (0.5, 0.5),
        (0., 0.),
    ]) * 2
    square_codes = [
        Path.MOVETO,
        Path.LINETO,
        Path.LINETO,
        Path.LINETO,
        Path.LINETO,
        Path.CLOSEPOLY,
    ]
    path = Path(square_verts * scale, square_codes)
    trans = matplotlib.transforms.Affine2D().translate(x, y)
    t_path = path.transformed(trans)
    patch = patches.PathPatch(
        t_path, facecolor=color.value, lw=line_weight, zorder=2)
    a = ax.add_patch(patch)
    ma = MonosaccharidePatch(saccharide_shape=(a,))
    return ma


def draw_triangle(ax, x, y, color, scale=0.1):
    path = Path(Path.unit_triangle().vertices * scale, Path.unit_triangle().codes)
    trans = matplotlib.transforms.Affine2D().translate(
        x, y).rotate_deg_around(x, y, -90)
    t_path = path.transformed(trans)
    patch = patches.PathPatch(
        t_path, facecolor=color.value, lw=line_weight, zorder=2)
    a = ax.add_patch(patch)
    ma = MonosaccharidePatch(saccharide_shape=(a,))
    return ma


def line_to(ax, sx, sy, ex, ey, zorder=1, color=None, lw=1):
    if color is None:
        color = default_line_color
    vertices = [
        (sx, sy),
        (ex, ey),
        (0, 0)
    ]
    codes = [
        Path.MOVETO,
        Path.LINETO,
        Path.STOP
    ]
    path = Path(vertices, codes)
    patch = patches.PathPatch(path, color=color, lw=lw, zorder=zorder)
    ax.add_patch(patch)
    return patch


def line_along(ax, points, zorder=1, color=None, lw=1):
    if color is None:
        color = default_line_color
    vertices = list(points)
    vertices.append([0, 0])
    codes = [Path.MOVETO] + [Path.LINETO] * (len(points) - 1)
    codes.append(Path.STOP)
    path = Path(vertices, codes)
    patch = patches.PathPatch(path, edgecolor=color, facecolor='none', lw=lw, zorder=zorder)
    ax.add_patch(patch)
    return patch


def transform_around_coords(x, y, xorigin, yorigin, transform):
    xy2 = np.array((x, y))
    xy1 = np.array((xorigin, yorigin))
    delta = xy1 - xy2
    reloc = transform.transform(delta)
    new_pos = xy1 + reloc
    return new_pos


class SymbolicNomenclatureBase(object):
    default_line_color = 'black'

    def get_layout_transform(self, orientation, **kwargs):
        if orientation in {'h', 'horizontal'}:
            return NodeLayoutTransforms(Affine2D().rotate_deg_around(0, 0, 90),
                                        Affine2D().rotate_deg(-90).translate(0, 0.15),
                                        Affine2D())

        elif orientation in {'v', 'vertical'}:
            return NodeLayoutTransforms(Affine2D().rotate_deg_around(0, 0, 0),
                                        Affine2D().rotate_deg(180).translate(0.05, -0.05),
                                        Affine2D())
        else:
            raise ValueError("Unknown Layout %r" % (orientation,))

    def draw(self, monosaccharide, x, y, ax, tree_node=None, scale=0.1, annotation_transform=None, **kwargs):
        raise NotImplementedError()

    def _get_horizontal_alignment(self, kwargs):
        value = kwargs.get("ha")
        if value is not None:
            return value
        value = kwargs.get("horizontal_alignment")
        return value

    def _get_vertical_alignment(self, kwargs):
        value = kwargs.get("va")
        if value is not None:
            return value
        value = kwargs.get("vertical_alignment")
        return value

    def draw_text(self, ax, x, y, text, scale=0.1, center=False, **kwargs):
        kwargs.setdefault('lw', line_weight / 20.0)
        fs = kwargs.get("fontsize", 2) * scale * .25
        t_path = TextPath((x, y), s=text, size=fs)
        ha = self._get_horizontal_alignment(kwargs)
        if center or ha == 'center':
            cx, cy = centroid(t_path)
            tf = Affine2D()
            tf.translate(x - cx, y - cy)
            t_path = t_path.transformed(tf)
        va = self._get_vertical_alignment(kwargs)
        if va == 'bottom':
            bbox = bounding_box(t_path)
            ymin = bbox.y0
            cx, cy = centroid(t_path)
            tf = Affine2D()
            tf.translate(0, cy - ymin)
            t_path = t_path.transformed(tf)
        patch = patches.PathPatch(t_path, facecolor="black", lw=kwargs['lw'], zorder=4)
        a = ax.add_patch(patch)
        return (a,)

    def format_text(self, text):
        label = ''.join([token.capitalize()[:2] for token in text.split('_', 1)])
        return label

    def line_to(self, ax, sx, sy, ex, ey, zorder=1, color=None, lw=1):
        if color is None:
            color = self.default_line_color
        vertices = [
            (sx, sy),
            (ex, ey),
            (0, 0)
        ]
        codes = [
            Path.MOVETO,
            Path.LINETO,
            Path.STOP
        ]
        path = Path(vertices, codes)
        patch = patches.PathPatch(path, color=color, lw=lw, zorder=zorder)
        ax.add_patch(patch)
        return patch

    def draw_substituents(self, ax, substituents, node_x, node_y, annotation_transform=None, **kwargs):
        subs = []
        sub_y = node_y - (0.15 * (len(substituents) - 1))
        sub_x = node_x - 0.45
        if annotation_transform is None:
            annotation_transform = Affine2D()
        for pos, subst_name in substituents:
            sub_x_tr, sub_y_tr = transform_around_coords(sub_x, sub_y, node_x, node_y, annotation_transform)
            sub_t = self.draw_text(
                ax, sub_x_tr, sub_y_tr, str(pos) + self.format_text(subst_name), center=True, fontsize=6)
            sub_y += 0.4
            subs.append(sub_t)
        return subs



import sys
import pdb
import functools
import traceback
import unittest
import hjson
import itertools
from collections import defaultdict

import glypy
from glypy.structure import constants, substituent, glycan, monosaccharide
from glypy.structure import link, named_structures, structure_composition
from glypy.structure import crossring_fragments
from glypy.io import glycoct
from glypy.io.nomenclature import identity, synonyms
from glypy.utils import StringIO, identity as ident_op, multimap, pickle, ET, enum
from glypy.composition import Composition, composition_transform
from glypy.algorithms import subtree_search
from glypy.algorithms import similarity

from .common import emit, load

ReducedEnd = monosaccharide.ReducedEnd
Substituent = glypy.Substituent
Glycan = glycan.Glycan


def debug_on(*exceptions):
    if not exceptions:
        exceptions = (AssertionError, )

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except exceptions:
                info = sys.exc_info()
                traceback.print_exception(*info)
                pdb.post_mortem(info[2])
        return wrapper
    return decorator
monosaccharides = named_structures.monosaccharides
glycans = named_structures.glycans

# monosaccharide_masses = json.load(open("./test_data/monosaccharide_masses.json"))
monosaccharide_structures = hjson.load(
    open("./glypy/structure/data/monosaccharides.hjson"))

wiki_masses = {
    "Iduronic Acid": 194.04,
    "Bacillosamine": 162.10,
    "Allose": 180.06,
}

#: Not common in any way other than
#: reused in many tests
common_glycan = '''
RES
1b:b-dglc-HEX-1:5
2b:b-dgal-HEX-1:5
3b:b-dglc-HEX-1:5
4s:n-acetyl
5b:a-lgal-HEX-1:5|6:d
6b:b-dgal-HEX-1:5
7b:b-dglc-HEX-1:5
8s:n-acetyl
9b:a-lgal-HEX-1:5|6:d
10b:b-dgal-HEX-1:5
LIN
1:1o(4+1)2d
2:2o(3+1)3d
3:3d(2+1)4n
4:3o(3+1)5d
5:3o(4+1)6d
6:6o(3+1)7d
7:7d(2+1)8n
8:7o(3+1)9d
9:7o(4+1)10d'''

branchy_glycan = '''
RES
1b:x-dglc-HEX-x:x
2s:n-acetyl
3b:b-dman-HEX-1:5
4b:a-dman-HEX-1:5
5b:b-dglc-HEX-1:5
6s:n-acetyl
7b:b-dgal-HEX-1:5
8b:b-dglc-HEX-1:5
9s:n-acetyl
10b:b-dgal-HEX-1:5
11b:a-dman-HEX-1:5
12b:b-dglc-HEX-1:5
13s:n-acetyl
14b:b-dgal-HEX-1:5
LIN
1:1d(2+1)2n
2:1o(4+1)3d
3:3o(3+1)11d
4:3o(6+1)4d
5:4o(2+1)8d
6:4o(6+1)5d
7:5d(2+1)6n
8:5o(4+1)7d
9:8d(2+1)9n
10:8o(4+1)10d
11:11o(2+1)12d
12:12d(2+1)13n
13:12o(4+1)14d'''

broad_n_glycan = '''
RES
1b:b-dglc-HEX-1:5
2s:n-acetyl
3b:b-dglc-HEX-1:5
4s:n-acetyl
5b:b-dman-HEX-1:5
6b:a-dman-HEX-1:5
7b:b-dglc-HEX-1:5
8s:n-acetyl
9b:b-dgal-HEX-1:5
10b:a-lgal-HEX-1:5|6:d
11b:b-dglc-HEX-1:5
12s:n-acetyl
13b:b-dgal-HEX-1:5
14b:a-dman-HEX-1:5
15b:b-dglc-HEX-1:5
16s:n-acetyl
17b:b-dgal-HEX-1:5
18b:b-dglc-HEX-1:5
19s:n-acetyl
20b:b-dgal-HEX-1:5
LIN
1:1d(2+1)2n
2:1o(4+1)3d
3:3d(2+1)4n
4:3o(4+1)5d
5:5o(3+1)14d
6:5o(6+1)6d
7:6o(2+1)11d
8:6o(6+1)7d
9:7d(2+1)8n
10:7o(3+1)10d
11:7o(4+1)9d
12:11d(2+1)12n
13:11o(4+1)13d
14:14o(2+1)18d
15:14o(4+1)15d
16:15d(2+1)16n
17:15o(4+1)17d
18:18d(2+1)19n
19:18o(4+1)20d'''

sulfated_glycan = '''
RES
1b:o-dgal-HEX-0:0|1:aldi
2b:b-dglc-HEX-1:5
3s:n-acetyl
4b:b-dgal-HEX-1:5
5b:b-dglc-HEX-1:5
6s:n-acetyl
7b:b-dgal-HEX-1:5
8b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
9s:n-acetyl
10s:sulfate
11s:sulfate
12s:sulfate
LIN
1:1o(3+1)2d
2:2d(2+1)3n
3:2o(4+1)4d
4:4o(3+1)5d
5:5d(2+1)6n
6:5o(4+1)7d
7:7o(6+2)8d
8:8d(5+1)9n
9:5o(6+1)10n
10:4o(6+1)11n
11:2o(6+1)12n
'''

complex_glycan = '''
RES
1b:x-dglc-HEX-1:5
2s:n-acetyl
3b:b-dglc-HEX-1:5
4s:n-acetyl
5b:b-dman-HEX-1:5
6b:a-dman-HEX-1:5
7b:b-dglc-HEX-1:5
8s:n-acetyl
9b:a-lgal-HEX-1:5|6:d
10b:b-dgal-HEX-1:5
11b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
12s:n-glycolyl
13b:b-dglc-HEX-1:5
14s:n-acetyl
15b:b-dgal-HEX-1:5
16s:n-acetyl
17b:b-dglc-HEX-1:5
18s:n-acetyl
19b:a-dman-HEX-1:5
20b:b-dglc-HEX-1:5
21s:n-acetyl
22b:a-lgal-HEX-1:5|6:d
23b:b-dgal-HEX-1:5
24b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
25s:n-glycolyl
26b:b-dglc-HEX-1:5
27s:n-acetyl
28b:a-lgal-HEX-1:5|6:d
29b:b-dgal-HEX-1:5
30b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
31s:n-acetyl
32b:a-lgal-HEX-1:5|6:d
LIN
1:1d(2+1)2n
2:1o(4+1)3d
3:3d(2+1)4n
4:3o(4+1)5d
5:5o(3+1)6d
6:6o(2+1)7d
7:7d(2+1)8n
8:7o(3+1)9d
9:7o(4+1)10d
10:10o(3+2)11d
11:11d(5+1)12n
12:6o(4+1)13d
13:13d(2+1)14n
14:13o(4+1)15d
15:15d(2+1)16n
16:5o(4+1)17d
17:17d(2+1)18n
18:5o(6+1)19d
19:19o(2+1)20d
20:20d(2+1)21n
21:20o(3+1)22d
22:20o(4+1)23d
23:23o(3+2)24d
24:24d(5+1)25n
25:19o(6+1)26d
26:26d(2+1)27n
27:26o(3+1)28d
28:26o(4+1)29d
29:29o(3+2)30d
30:30d(5+1)31n
31:1o(6+1)32d
'''


class TestNamedStructure(unittest.TestCase):

    def test_accessors(self):
        self.assertEqual(named_structures.monosaccharides.Fucose,
                         named_structures.monosaccharides["Fucose"])
        self.assertNotEqual(named_structures.monosaccharides.Man,
                            named_structures.monosaccharides.Hex)
        self.assertEqual(named_structures.glycans["N-Linked Core"],
                         named_structures.glycans["N-Linked Core"])
        self.assertEqual(named_structures.motifs["N-Glycan hybrid 1"],
                         named_structures.motifs["N-Glycan hybrid 1"])
        self.assertIn(
            named_structures.motifs["N-Glycan core basic 1"],
            named_structures.motifs.motif_class("N-Glycan").values())

        self.assertIn(
            named_structures.motifs['Beta-glucan'],
            named_structures.motifs.motif_category("full").values())


class TestStructureComposition(unittest.TestCase):

    def test_missing_composition(self):
        import warnings
        warnings.filterwarnings('ignore')
        self.assertEqual(
            {}, structure_composition.monosaccharide_composition['hexadodecose'])
        warnings.filterwarnings('always')


class TestCrossRing(unittest.TestCase):

    def test_unroll_ring(self):
        linear = crossring_fragments.unroll_ring(monosaccharides.GlcNAc)
        for i in range(len(linear)):
            self.assertEqual(linear[i]['backbone'], Composition("CH2O"))
            if i == 1:
                self.assertTrue(len(linear[i]['substituent_links']) > 0)

    class FragmentRecord(object):
        def __init__(self, name, mass, permethylated_mass, cleave_1, cleave_2, kind):
            self.name = name
            self.mass = float(mass)
            self.permethylated_mass = float(permethylated_mass)
            self.cleave_1 = int(cleave_1)
            self.cleave_2 = int(cleave_2)
            self.kind = kind
            self.composition = {}

        def __repr__(self):
            return "FragmentRecord({name} {mass})".format(**self.__dict__)

    def load_all(self):
        tree = ET.parse('./test_data/Cross-ring-data.xml')
        sugars = {}
        by_name = {}
        for cls in tree.iterfind(".//class"):
            sugar_class = {}
            for case in cls.iterfind(".//sugar"):
                isomers = [tag.attrib['abbr'] for tag in case.findall(".//isomer")]
                frags = defaultdict(dict)
                for frag in case.iterfind(".//fragment"):
                    data = [frag.attrib[n] for n in ["name", "mass_mono", "mass_pm_mono", "cleav1", "cleav2", "type"]]
                    rec = self.FragmentRecord(*data)
                    rec.composition = dict(map(
                        lambda x: (x.attrib['element'], x.attrib['count']), frag.findall(".//composition")))
                    frags[rec.cleave_1, rec.cleave_2][rec.kind] = rec
                sugar_class[case.attrib['subclass']] = frags
                for name in isomers:
                    by_name[name] = frags
            sugars[cls.attrib['name']] = sugar_class
        return by_name

    def test_monosacchide_crossring(self):
        test_cases = [
            'Glc',
            "GlcNAc",
            "Xyl",
            'GalA',
            'Fuc',
            'IdoA',
            "KDN"
        ]
        store = self.load_all()
        for case in test_cases:
            test = store[case]
            target = monosaccharides[case]
            for k, v in test.items():
                target_d = {t.kind: t for t in crossring_fragments.crossring_fragments(target, k[0], k[1])}
                target_d_permethylated = {t.kind: t for t in crossring_fragments.crossring_fragments(
                    composition_transform.derivatize(target.clone(), "methyl"), k[0], k[1])}
                for kind in {"A", "X"}:
                    self.assertAlmostEqual(v[kind].mass, target_d[kind].mass(), 3)
                    self.assertEqual(
                        target_d[kind].open_attachment_sites(),
                        target_d[kind].clone().open_attachment_sites())
                    self.assertAlmostEqual(v[kind].permethylated_mass, target_d_permethylated[kind].mass(), 3)


class TestSubstituent(unittest.TestCase):

    def test_substituent_substituent(self):
        parent = substituent.Substituent("n-acetyl")
        pmass = parent.mass()
        child = substituent.Substituent("methyl")
        cmass = child.mass()

        parent.add_substituent(child, 2)
        self.assertAlmostEqual(
            parent.mass(), (pmass + cmass - Composition(H=2).mass))
        self.assertEqual(
            parent.total_composition(), parent.composition + child.composition)
        self.assertEqual(parent.children()[0][1], child)
        self.assertEqual(child.parents()[0][1], parent)
        parent.drop_substituent(2)
        self.assertAlmostEqual(parent.mass(), pmass)

    def test_equality(self):
        sub_1 = substituent.Substituent("n-acetyl")
        sub_2 = substituent.Substituent("n-acetyl")
        self.assertEqual(sub_1, sub_2)
        parent = named_structures.monosaccharides['Hex']
        parent.add_substituent(sub_1, 3)
        self.assertNotEqual(sub_1, sub_2)

    def test_clone(self):
        parent = substituent.Substituent("n-acetyl")
        child = substituent.Substituent("methyl")

        parent.add_substituent(child, 2)

        dup = parent.clone()
        self.assertEqual(parent, dup)
        self.assertNotEqual(child, child.clone())

    def test_derivatize_pathway_setter(self):
        substituent.Substituent("not-real", None, Composition("H2O"), can_nh_derivatize=False, is_nh_derivatizable=True)
        x = substituent.Substituent("not-real", None, Composition("H2O"))
        self.assertTrue(x.is_nh_derivatizable)
        substituent.DerivatizePathway.register("not-real", False, False)
        substituent.Substituent("not-real", None, Composition("H2O"))
        x = substituent.Substituent("not-real", None, Composition("H2O"))
        self.assertFalse(x.is_nh_derivatizable)


class TestMultiMap(unittest.TestCase):

    def test_iterators(self):
        from collections import Counter
        mm = multimap.MultiMap(a=1, b=3)
        mm['a'] = 3
        self.assertTrue(set(mm.keys()) == {'a', 'b'})
        self.assertTrue(set(mm.items()) == {('a', 1), ('a', 3), ('b', 3)})
        self.assertTrue(Counter(mm.values()) == Counter({1: 1, 3: 2}))

    def test_ordered_multimap(self):
        mm = multimap.MultiMap(a=1, b=3)
        mm['a'] = 3
        omm = multimap.OrderedMultiMap(a=1, b=3)
        omm['a'] = 3
        self.assertEqual(mm, omm)
        omm['c'] = 1
        self.assertNotEqual(mm, omm)
        self.assertNotEqual(omm, mm)
        self.assertFalse('c' in mm)

    def test_clone(self):
        omm = multimap.OrderedMultiMap(a=1, b=3)
        alt = omm.clone()
        self.assertEqual(omm, alt)
        omm['a'] = 2
        self.assertNotEqual(omm, alt)

    def test_has_value(self):
        mm = multimap.MultiMap(a=1, b=3)
        self.assertTrue(mm.has_value(1))
        self.assertFalse(mm.has_value('a'))
        mm.popv(1)
        self.assertFalse(mm.has_value(1))


class TestConstant(unittest.TestCase):

    def test_translate(self):
        self.assertTrue(
            constants.Modification.d == constants.Modification['d'])
        self.assertTrue(constants.Modification.d == constants.Modification[
                        constants.Modification.d.value])

    def test_compare(self):
        self.assertTrue(constants.Modification.d == 'd')
        self.assertTrue(
            constants.Modification.d == constants.Modification.d.value)
        self.assertNotEqual(
            constants.Modification.d, constants.Stem[constants.Modification.d.value])
        self.assertRaises(KeyError, lambda: constants.SuperClass[1])
        self.assertNotEqual(constants.Modification.d, constants.Modification.a)

    def test_multiname(self):
        Modification = constants.Modification
        self.assertTrue(Modification.d == 'd')
        self.assertFalse(Modification.d == 'Deoxidation')
        # Add new name to existing EnumValue
        Modification.d.add_name("Deoxidation")
        self.assertTrue(Modification.d == 'Deoxidation')

        # Add new name to class by re-using EnumValue
        self.assertFalse(Modification.d == "deoxy")
        Modification.deoxy = Modification.d
        self.assertTrue(Modification.d == "deoxy")
        self.assertEqual(Modification.deoxy, Modification.Deoxidation)

        # Add new name to class and EnumValue by interpolating value
        self.assertFalse(Modification.deoxy == 'deoxy2')
        Modification.deoxy2 = Modification.d.value
        self.assertTrue(Modification.deoxy == 'deoxy2')
        self.assertRaises(KeyError, lambda: Modification.d.add_name('a'))

    def test_wrap_value(self):
        constants.Modification.not_a_real_modification = -5
        self.assertTrue(isinstance(constants.Modification.not_a_real_modification, enum.EnumValue))
        self.assertEqual(constants.Modification.not_a_real_modification.resolve(
            constants.Modification), -5)

    def test_resolve_failure(self):
        mapping = {'d': 5}
        self.assertTrue(constants.Modification.d.resolve(mapping) == 5)
        self.assertRaises(KeyError, lambda: constants.Modification.aldi.resolve(mapping))

    def test_instantiate_error(self):
        self.assertRaises(Exception, enum.Enum)

    def test_crossclass_inequality(self):
        class E1(enum.Enum):
            A = 1

        class E2(enum.Enum):
            A = 1
        self.assertNotEqual(E1.A, E2.A)


# class IdentifyTests(unittest.TestCase):

#     def test_is_a_predicate(self):
#         for name, mono in monosaccharides.items():
#             result = identity.is_a(mono, name)
#             self.assertTrue(result)

#     def test_identify_as(self):
#         for name, mono in monosaccharides.items():
#             if name in {"Hex", "Pen", "Oct", "Hep", "Non"}:
#                 continue
#             pref_name = identity.identify(mono)
#             if not (name == pref_name or name in synonyms.monosaccharides[pref_name]):
#                 raise AssertionError(
#                     "{}".format((name, pref_name, synonyms.monosaccharides[pref_name])))

#     def test_identify_substituents(self):
#         self.assertTrue(
#             identity.is_a(Substituent("n-acetyl"), Substituent("n-acetyl")))
#         self.assertFalse(
#             identity.is_a(Substituent('methyl'), Substituent('n-acetyl')))
#         self.assertFalse(
#             identity.is_a(monosaccharides.Man, Substituent('n-acetyl')))
#         self.assertFalse(
#             identity.is_a(Substituent('n-acetyl'), monosaccharides.Man))

#     def test_get_preferred_name(self):
#         self.assertTrue(identity.get_preferred_name('bdMan') == 'Man')

#     def test_identify_failure(self):
#         # Will fail because Hex is blacklisted
#         self.assertRaises(
#             identity.IdentifyException, lambda: identity.identify(monosaccharides.Hex))

#     def test_precision(self):
#         self.assertFalse(identity.is_a(monosaccharides.Kdn, monosaccharides.NeuAc))
#         self.assertFalse(identity.is_a(monosaccharides.NeuAc, monosaccharides.Kdn))

#     def test_grouping_axes(self):
#         tree = identity.residue_list_to_tree((monosaccharides).values())


if __name__ == '__main__':
    unittest.main()



from glypy.utils import enum, root as rootp, uid
from glypy.structure import monosaccharide


class StructurePrecisionEnum(enum.Enum):
    unknown = -1
    ranging = 1
    exact = 2


class AbstractGraphEntryEnum(enum.Enum):
    parent = 1
    child = 2
    internal = 3


def try_int(v):
    try:
        return int(v)
    except Exception:
        return None


def decorate_tree(tree, decorate_value):
    '''
    Transform each node in the tree's id value
    into a tuple of `(decorate_value, node.id)`
    to use common ids across identical graphs but
    discriminable between attached copies.

    Parameters
    ----------
    tree: Glycan
    decorate_value: int

    Returns
    -------
    Glycan
    '''
    for node in tuple(tree):
        node.id = (decorate_value, node.id)
        for i, substituent in node.substituents():
            substituent.id = (decorate_value, substituent.id)
    return tree


def decorated_value(tree):  # pragma: no cover
    '''
    Get the decorating value from a tree's root node.id.

    Returns `None` if the tree is undecorated

    Parameters
    ----------
    tree: Glycan

    Returns
    -------
    int or None
    '''
    try:
        return next(iter(tree.root.id))
    except Exception:
        return None


def get_decorated(tree, id):  # pragma: no cover
    '''
    As :meth:`Glycan.get`, but with awareness of decorated
    node.id attributes. Will fall back to the normal get method
    if the tree is undecorated.

    Parameters
    ----------
    tree: Glycan
    id: int

    Returns
    -------
    Monosaccharide
    '''
    d = decorated_value(tree)
    if d is None:
        return tree.get(id)
    else:
        return tree.get((d, id))


def undecorate_tree(tree):
    '''
    Remove decoration from a tree and reindex its nodes.

    Depends upon  :attr:`Glycan.index` to find nodes

    Parameters
    ----------
    tree: Glycan

    Returns
    -------
    Glycan
    '''
    i = 1
    for node in list(tree):
        node.id = i
        i += 1
        for j, substituent in node.substituents():
            substituent.id = i
            i += 1

    # tree.reindex()
    return tree


def find_root(tree):
    root = rootp(tree)

    while True:
        parents = root.parents()
        if not parents:
            break
        root = parents[0][1]
    return root


class NodeCollection(object):

    @classmethod
    def from_node(cls, root):
        nodes = []
        for node in monosaccharide.traverse(root):
            nodes.append(node)
            try:
                for _, subst in node.substituents():
                    nodes.append(subst)
            except AttributeError:
                pass
        return cls(nodes)

    def __init__(self, nodes):
        self.nodes = list(nodes)
        self.root = self.nodes[0]

    def __iter__(self):
        return monosaccharide.traverse(self.root)

    def deindex(self):
        base = uid()
        for i, node in enumerate(self.nodes):
            node.id += base
            node.id *= -1
        return self

    def reindex(self):
        for i, node in enumerate(self.nodes):
            node.id = i + 1
        return self

    def clone(self, *args, **kwargs):
        root = monosaccharide.graph_clone(self.root)
        return self.from_node(root)

    def get(self, key):
        for node in self.nodes:
            if node.id == key:
                return node
        raise IndexError(key)

    def __getitem__(self, i):
        return self.get(i)

    def __root__(self):
        return self.root



import string
from glypy.io.file_utils import ParserError


def base52(x):
    code = []
    if x == 0:
        return string.ascii_letters[0]
    while x > 0:
        i = x % 52
        code.append(i)
        x //= 52
    code = code[::-1]
    n = len(code)
    if n == 1:
        return ''.join([string.ascii_letters[c] for j, c in enumerate(code)])
    else:
        return ''.join([string.ascii_letters[c - 1 if j != n - 1 else c] for j, c in enumerate(code)])


class WURCSError(ParserError):
    pass


class WURCSFeatureNotSupported(WURCSError):
    pass



import logging
import json
from collections import defaultdict

logger = logging.getLogger(__name__)


def _str_dump_multimap(mm):  # pragma: no cover
    d = defaultdict(list)
    for k, v in mm.items():
        d[k].append(repr(v))
    return json.dumps(d, sort_keys=True, indent=4)


class MultiMap(object):
    __slots__ = ['contents', 'clean']

    '''Implements a simple MultiMap data structure on top of a dictionary of lists'''
    def __init__(self, **kwargs):
        self.contents = defaultdict(list)
        for k, v in kwargs.items():
            self.contents[k].append(v)
        self.invalidate()

    def invalidate(self):
        self.clean = False

    def __getitem__(self, key):
        return self.contents[key]

    def __setitem__(self, key, value):
        self.contents[key].append(value)
        self.invalidate()

    def pop(self, key, value):
        '''
        Removes `value` from the collection of values stored at `key`
        and returns the `tuple` `(key, value)`

        Raises
        ------
        IndexError
        KeyError
        '''
        objs = self.contents[key]
        objs.pop(objs.index(value))
        if len(objs) == 0:
            self.contents.pop(key)
        self.invalidate()
        return len(objs)

    def popv(self, value):
        for k in self:
            if value in self[k]:
                self.pop(k, value)
                return len(self[k])
        self.invalidate()
        return None

    def __iter__(self):
        return iter(self.contents)

    def __contains__(self, key):
        return key in self.contents

    def keys(self):
        '''
        Returns an iterator over the keys of :attr:`contents`
        An alias of :meth:`__iter__`
        '''
        return iter(self)

    def values(self):
        '''
        Returns an iterator over the values of :attr:`contents`
        '''
        for k in self:
            for v in self[k]:
                yield v

    def items(self):
        '''
        Returns an iterator over the items of :attr:`contents`. Each item
        takes the form of `(key, value)`.
        '''
        for k in self:
            for v in self[k]:
                yield (k, v)

    def lists(self):
        return self.contents.items()

    def __len__(self):
        '''
        Returns the number of items in :attr:`contents`
        '''
        return sum(len(self[k]) for k in self)

    def __repr__(self):  # pragma: no cover
        return _str_dump_multimap(self)

    def __eq__(self, other):
        if other is None:  # pragma: no cover
            return False
        for a in self.keys():
            if a in other:
                if self[a] != other[a]:
                    return False
            else:
                if self[a] != []:
                    return False
        for b in other.keys():
            if b in self:
                continue
            else:
                if other[b] != []:
                    return False
        # for a, b in izip_longest(self.items(), other.items()):
        #     if a != b:
        #         return False
        return True

    def __ne__(self, other):
        return not self == other

    def update(self, mapping):
        for k, v in mapping.items():
            self[k] = v
        self.invalidate()

    def has_value(self, value):
        for v in self.values():
            if v == value:
                return True
        return False

    def __reduce__(self):
        return self.__class__, (), self.__getstate__()

    def __getstate__(self):
        return self.contents

    def __setstate__(self, state):
        self.contents = state
        self.invalidate()

    def copy(self):
        new = self.__class__()
        for k, v in self.items():
            new[k].extend(v)
        return new

    def clone(self):
        return self.copy()


class OrderedMultiMap(MultiMap):
    '''
    Implements a simple MultiMap data structure on top of a dictionary of lists
    that remembers the order keys were first inserted in.
    '''

    __slots__ = ['key_order', ]

    def __init__(self, **kwargs):
        self.contents = defaultdict(list)
        self.key_order = []
        for k, v in kwargs.items():
            if k not in self.key_order:
                self.key_order.append(k)
            self.contents[k].append(v)
        self.invalidate()

    def __iter__(self):
        '''
        Returns an iterator over the keys of :attr:`contents` in the order
        they were added.
        '''
        for key in self.key_order:
            yield key

    #: Alias of :meth:`__iter__`
    keys = __iter__

    def values(self):
        for key in self.key_order:
            for v in self[key]:
                yield v

    def items(self):
        '''
        As in :class:`MultiMap`, but items are yielded in the order their keys were
        first added.
        '''
        for key in self.key_order:
            for v in self[key]:
                yield (key, v)

    def lists(self):
        for key in self.key_order:
            yield key, self[key]

    def reorder(self, new_order):
        assert set(new_order) == set(self.key_order)
        self.key_order = list(new_order)

    def __setitem__(self, key, value):
        if key not in self.key_order:
            self.key_order.append(key)
        self.contents[key].append(value)
        self.invalidate()

    def __repr__(self):  # pragma: no cover
        return ''.join((repr(self.key_order), '\n', _str_dump_multimap(self)))

    def __getstate__(self):
        return self.contents, self.key_order

    def __setstate__(self, state):
        self.contents, self.key_order = state
        self.invalidate()



'''

'''

from .base import ModificationBase
from glypy.composition.structure_composition import modification_compositions
from glypy.io.format_constants_map import modification_map
from glypy.utils import uid

glycoct_map = {v: k for k, v in modification_map.items()}


def resolve_composition_rule(name):
    try:
        return modification_compositions[name]
    except KeyError:
        return modification_compositions[glycoct_map[name]]


class Modification(ModificationBase):
    __slots__ = ('name', 'position', 'composition', 'id')

    def __init__(self, name, position, composition=None, id=None):
        composition = composition or resolve_composition_rule(name)(position)
        if id is None:
            id = uid()
        self.name = name
        self.position = position
        self.composition = composition
        self.id = id

    def __repr__(self):  # pragma: no cover
        return "<Modification {name}@{position}>".format(
            name=self.name, position=self.position)

    def __eq__(self, other):
        return (self.name == other) or\
            (self.name == other.name and self.composition == other.composition)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.name)

    def to_glycoct(self):
        return glycoct_map.get(self.name, self.name)

    def clone(self, prop_id=True):
        return self.__class__(self.name, self.position, self.composition, id=self.id if prop_id else None)

    def __reduce__(self):
        return self.__class__, (self.name, self.position, self.composition, self.id)

    def site_count(self):
        return 1


class MultiSiteModification(Modification):
    __slots__ = ('positions',)

    def __init__(self, name, positions, composition=None, id=None):
        super(MultiSiteModification, self).__init__(
            name, positions[0], composition=composition, id=id)
        self.positions = tuple(positions)

    def clone(self, prop_id=True):
        return self.__class__(self.name, self.positions, self.composition, id=self.id if prop_id else None)

    def __reduce__(self):
        return self.__class__, (self.name, self.positions, self.composition, self.id)

    def site_count(self):
        return len(self.positions)

    def __repr__(self):  # pragma: no cover
        return "<MultiSiteModification {name}@{positions}>".format(
            name=self.name, positions=self.positions)




from ..utils import make_counter
from ..structure.base import SaccharideCollection, SaccharideBase, SubstituentBase
from ..structure import Substituent
from ..structure import Monosaccharide
from ..structure import Modification
from .composition import Composition
from six import string_types as basestring


def derivatize(saccharide, substituent):
    r'''
    For each monosaccharide and viable substituent, attach the specified
    substituent.

    .. warning::
        This function mutates the input `saccharide` object, because copying objects
        is expensive. If you want to continue using the underivatized object, create
        a copy of it using the object's `clone` method.

    When called on an instance of :class:`SaccharideCollection`, such as |Glycan| or
    :class:`GlycanComposition`, that type's `_derivatized` method will be called for
    any special book-keeping that must be done. This is a NoOp for |Glycan|

    See Also
    --------
    :func:`strip_derivitization`

    '''
    if isinstance(substituent, basestring):
        substituent = Substituent(substituent)
    id_base = make_counter(-substituent.id * 32)
    if isinstance(saccharide, SaccharideCollection):
        for node in saccharide:
            if node.node_type is SaccharideBase.node_type:
                _strip_derivatization_monosaccharide(node)
                _derivatize_monosaccharide(node, substituent, id_base)
            elif node.node_type is SubstituentBase.node_type:
                _strip_derivatization_substituent(node)
                _derivatize_substituent(node, substituent, id_base)
        saccharide._derivatized(substituent, id_base)
    elif isinstance(saccharide, SaccharideBase):
        _derivatize_monosaccharide(saccharide, substituent, id_base)
    return saccharide


def _derivatize_monosaccharide(monosaccharide_obj, substituent, id_base=None):
    '''
    The internal worker for :func:`derivatize`. It should not be called directly.

    Adds a copy of `substituent` to every open position on `monosaccharide_obj` and
    its substituents and reducing end

    Parameters
    ----------
    monosaccharide_obj: Monosaccharide
    substituent: Substituent
    id_base: function or None

    Returns
    -------
    monosaccharide_obj
    '''
    if id_base is None:
        id_base = make_counter(-substituent.id * 32)
    attachment_composition_loss = substituent.attachment_composition_loss()
    open_sites, unknowns = monosaccharide_obj.open_attachment_sites()
    for site in open_sites[unknowns:]:
        s = substituent.clone()
        s.id = id_base()
        s._derivatize = True
        monosaccharide_obj.add_substituent(
            s, parent_loss=attachment_composition_loss,
            position=site, child_loss=Composition(H=1), child_position=1)
    for p, subst in monosaccharide_obj.substituents():
        _derivatize_substituent(subst, substituent, id_base)
    red_end = monosaccharide_obj.reducing_end
    if red_end is not None:
        _derivatize_reducing_end(red_end, substituent, id_base)

    for pos, mod in monosaccharide_obj.modifications.items():
        if mod == Modification.a:
            s = substituent.clone()
            s._derivatize = True
            s.id = id_base()
            monosaccharide_obj.add_substituent(
                s, position=pos, parent_loss=attachment_composition_loss, max_occupancy=3,
                child_loss=Composition(H=1), child_position=1)


def _derivatize_substituent(subst, deriv, id_base):
    if subst.is_nh_derivatizable and deriv.can_nh_derivatize:
        s = deriv.clone()
        s._derivatize = True
        s.id = id_base()
        subst.add_substituent(s, position=2, child_position=1)


def _derivatize_reducing_end(reducing_end, substituent, id_base=None):
    attachment_composition_loss = substituent.attachment_composition_loss()
    for i in range(1, reducing_end.valence + 1):
        s = substituent.clone()
        s._derivatize = True
        s.id = id_base()
        reducing_end.add_substituent(
            s, parent_loss=attachment_composition_loss, max_occupancy=3,
            position=i, child_loss=Composition(H=1), child_position=1)


def strip_derivatization(saccharide):
    '''
    For each monosaccharide and viable substituent, remove all substituents
    added by :func:`derivatize`.

    .. warning::
        This function, like :func:`derivatize`, will mutate the input `saccharide`.

    When called on an instance of :class:`SaccharideCollection`, such as |Glycan| or
    :class:`GlycanComposition`, that type's `_strip_derivatization` method will be called for
    any special book-keeping that must be done. This is a NoOp for |Glycan|

    See Also
    --------
    :func:`.derivatize`
    '''
    if isinstance(saccharide, SaccharideCollection):
        for node in saccharide:
            if node.node_type is SaccharideBase.node_type:
                _strip_derivatization_monosaccharide(node)
            elif node.node_type is SubstituentBase.node_type:
                _strip_derivatization_substituent(node)
        saccharide._strip_derivatization()
    else:
        _strip_derivatization_monosaccharide(saccharide)
    return saccharide


def _strip_derivatization_monosaccharide(monosaccharide_obj):
    for pos, subst_link in list(monosaccharide_obj.substituent_links.items()):
        if subst_link.child._derivatize:
            monosaccharide_obj.drop_substituent(pos, subst_link.child)
        else:
            sub_node = subst_link.child
            _strip_derivatization_substituent(sub_node)
    red_end = monosaccharide_obj.reducing_end
    if red_end is not None:
        _strip_derivatization_reducing_end(red_end)


def _strip_derivatization_substituent(sub_node):
    for sub_pos, subst_link in list(sub_node.links.items()):
        try:
            if subst_link.child._derivatize:
                sub_node.drop_substituent(sub_pos, subst_link.child)
        except AttributeError:
            if subst_link.child.node_type is Monosaccharide.node_type:
                _strip_derivatization_monosaccharide(subst_link.child)


def _strip_derivatization_reducing_end(reducing_end):
    for pos, subst_link in list(reducing_end.links.items()):
        if subst_link.child._derivatize:
            reducing_end.drop_substituent(pos, subst_link.child)


def has_derivatization(residue):
    if residue.node_type is Monosaccharide.node_type:
        for link in residue.substituent_links.values():
            if link.child._derivatize:
                return link.child
    return None


# WIP
class DerivatizeBase(object):  # pragma: no cover
    def __init__(self, substituent, white_list, black_list, *args, **kwargs):
        self.substituent = substituent
        self.white_list = white_list
        self.black_list = black_list

    def on_reducing_end(self, node):
        pass

    def on_terminal(self, node):
        pass

    def on_substituent(self, node):
        pass



from . import composition
from .composition import Composition, calculate_mass
from .base import formula, ChemicalCompositionError

__all__ = [
    "composition", "Composition", "calculate_mass",
    "formula", "ChemicalCompositionError",
    "composition_transform"
]



'''
A module for operating on GlycoMinds Linear Code

Assumes that the structure's root is the right-most residue, as shown in
:title-reference:`A Novel Linear Code Nomenclature for Complex Carbohydrates, Banin et al.`.

Currently does not handle the sigils indicating deviation from the common forms.
'''

import re
from collections import OrderedDict, deque

from six import string_types as basestring

from glypy.io import format_constants_map
from glypy.io.nomenclature import identity
from glypy.structure import constants, named_structures, Monosaccharide, Glycan, Substituent
from glypy.utils import invert_dict

from glypy.io.file_utils import ParserInterface, ParserError

Stem = constants.Stem
Configuration = constants.Configuration

# A static copy of monosaccharide names to structures for copy-free comparison
monosaccharide_reference = {k: v.clone() for k, v in named_structures.monosaccharides.items()}
# Unset the anomericity as this does not influence monosaccharide resolution
for k, v in monosaccharide_reference.items():
    v.anomer = None

#: A mapping from common monosaccharide names to their symbol, ordered by priority
monosaccharides_to = OrderedDict((
    ("Glc", 'G'),
    ("Gal", 'A'),
    ("GlcNAc", 'GN'),
    ("GalNAc", 'AN'),
    ("Man", "M"),
    ("Neu", "N"),
    ("NeuAc", "NN"),
    ("NeuGc", "NJ"),
    ("KDN", "K"),
    ("Kdo", "W"),
    ("GalA", "L"),
    ("IdoA", "I"),
    ("Rha", "H"),
    ("Fuc", "F"),
    ("Xyl", "X"),
    ("Rib", "B"),
    ("Ara", "R"),
    ("GlcA", "U"),
    ("All", 'O'),
    # ("Api", 'P'),
    ("Fru", "E")
))

#: A mapping from symbol to common monosaccharide name
monosaccharides_from = invert_dict(dict(monosaccharides_to))

#: A mapping from common substituent names to symbol
substituents_to = {
    'amino': 'Q',
    'ethanolominephosphate': 'PE',
    'inositol': "IN",
    'methyl': "ME",
    'n_acetyl': 'N',
    'o_acetyl': 'T',
    'phosphate': 'P',
    'phosphocholine': "PC",
    'pyruvate': 'PYR',
    'sulfate': 'S',
    'sulfide': 'SH',
    '2-aminoethylphosphonic acid': 'EP'
}

#: A mapping from symbol to common substituent names
substituents_from = invert_dict(substituents_to)


anomer_map_from = dict(format_constants_map.anomer_map)
anomer_map_from['?'] = anomer_map_from.pop('x')
anomer_map_to = invert_dict(anomer_map_from)


def get_relevant_substituents(residue):
    '''
    Retrieve the set of substituents not implicitly included
    in the base type's symbol name.
    '''
    positions = [p for p, sub in residue.substituents()]
    substituents = [sub.name for p, sub in residue.substituents()]
    if identity.is_a(residue, monosaccharide_reference["HexNAc"], exact=False) or\
       identity.is_a(residue, monosaccharide_reference["NeuAc"], exact=False):
        i = substituents.index("n_acetyl")
        substituents.pop(i)
        positions.pop(i)
    elif identity.is_a(residue, monosaccharide_reference["NeuGc"], exact=False):
        i = substituents.index("n_glycolyl")
        substituents.pop(i)
        positions.pop(i)
    return zip(substituents, positions)


def substituent_to_linear_code(substituent, position=None):
    '''
    Translate a |Substituent| to Linear Code. Include's the substituent's
    position if it is known.

    Parameters
    ----------
    substituents: Substituent or str
        The structure or the name to translate
    position:
        The position of the structure to try including

    Returns
    -------
    str

    Raises
    ------
    KeyError:
        When an unknown symbol is encountered
    '''
    symbol = None
    try:
        symbol = substituents_to[substituent.name]
    except:
        symbol = substituents_to[substituent]
    if position is not None:
        if position == -1:
            position = ''
        return "{}{}".format(position, symbol)
    else:
        return symbol


def monosaccharide_to_linear_code(monosaccharide, max_tolerance=3):
    '''
    Perform iteratively permissive attempts to translate `monosaccharide` into
    a nomenclature symbol.

    .. note::
        Uses a multi-pass approach. Could alternatively do a single pass
        and keep the best match.

    Parameters
    ----------
    monosaccharide: Monosaccharide
        The residue to be translated
    max_tolerance: int
        The maximum error tolerance to allow while looking for a match

    Returns
    -------
    str

    Raises
    ------
    ValueError:
        When no suitable translation can be found
    KeyError:
        When an unknown symbol is encountered
    '''
    tolerance = 0
    if identity.is_generic_monosaccharide(monosaccharide):
        raise LinearCodeError("Linear Code does not support generic monosaccharide %s" % str(monosaccharide))
    while tolerance <= max_tolerance:
        for k, v in monosaccharides_to.items():
            if k not in monosaccharide_reference:
                continue
            if identity.is_a(monosaccharide, monosaccharide_reference[k], tolerance=tolerance):
                residue_sym = v
                substituents_sym = [(
                    substituent_to_linear_code(
                        *s)) for s in get_relevant_substituents(monosaccharide)]
                if len(substituents_sym) > 0:
                    residue_sym = residue_sym + '[{}]'.format(', '.join(substituents_sym))
                residue_sym = residue_sym + anomer_map_to[monosaccharide.anomer]
                return residue_sym

        tolerance += 1
    raise LinearCodeError("Cannot map {} to Linear Code".format(monosaccharide))


def priority(sym):
    '''
    Calculate the branching priority for a given symbol or |Monosaccharide|.
    Used when deciding when a residue is considered branch or backbone.

    Parameters
    ----------
    sym: str or Monosaccharide

    Returns
    -------
    int
    '''
    i = 0
    is_str = isinstance(sym, basestring)
    for key, value in monosaccharides_to.items():
        i += 1
        if (identity.is_a(sym, monosaccharide_reference[key])) if not is_str else value == sym:
            return i
    return -1


def glycan_to_linear_code(structure=None, max_tolerance=3):
    '''
    Translate a |Glycan| structure into Linear Code. Called from :func:`to_linear_code`.
    Recursively operates on branches.

    Parameters
    ----------
    structure: Glycan or Monosaccharide
        The glycan to be translated. Translation starts from `glycan.root` if `structure`
        is a |Glycan|.
    max_tolerance: int
        The maximum amount of deviance to allow when translating |Monosaccharide| objects
        into nomenclature symbols

    Returns
    -------
    deque
    '''
    base = structure.root if isinstance(structure, Glycan) else structure
    stack = [(1, base)]
    outstack = deque()
    while(len(stack) > 0):
        outedge_pos, node = stack.pop()
        if outedge_pos in {1, -1}:
            outedge_pos = ''
        outstack.appendleft(monosaccharide_to_linear_code(node, max_tolerance=max_tolerance) + str(outedge_pos))
        children = []
        for pos, child in node.children():
            rank = priority(child)
            children.append((pos, child, rank))
        if len(children) > 1:
            ordered_children = sorted(children, key=lambda x: x[2])
            for pos, child, rank in ordered_children[:-1]:
                branch = '({branch}{attach_pos})'.format(
                    branch=''.join(glycan_to_linear_code(child, max_tolerance=max_tolerance)),
                    attach_pos=pos
                )
                outstack.appendleft(branch)
            pos, child, rank = ordered_children[-1]
            stack.append((pos, child))
        elif len(children) == 1:
            pos, child, rank = children[0]
            stack.append((pos, child))
    return outstack


def to_linear_code(structure):
    '''
    Translates `structure` to Linear Code.

    Parameters
    ----------
    structure: Monosaccharide or Glycan

    Returns
    -------
    str
    '''
    if isinstance(structure, Monosaccharide):
        return monosaccharide_to_linear_code(structure)
    else:
        return ''.join(list(glycan_to_linear_code(structure)))


class LinearCodeError(ParserError):
    pass


def monosaccharide_from_linear_code(residue_str, parent=None):
    '''
    Helper function for :func:`parse_linear_code`. Given a residue string
    of the form "(base_type)(anomer)(outedge?)([substituents]?)", construct
    a |Monosaccharide| object. If `parent` is not |None|, connect the
    resulting |Monosaccharide| to `parent` at `outedge`.
    '''
    base_type, substituents, anomer, outedge = re.search(r"([A-Z]+)(\[.*?\])?([abo\?]?)(.)?", residue_str).groups()
    base = named_structures.monosaccharides[monosaccharides_from[base_type]]
    base.anomer = anomer_map_from.get(anomer, None)
    if substituents is not None:
        for subst_str in substituents[1:-1].split(','):
            pos, name = re.search(r'(\d*)(.+)', subst_str).groups()
            subst_object = Substituent(substituents_from[name])
            try:
                pos = int(pos)
            except (ValueError, TypeError):
                pos = -1
            base.add_substituent(subst_object, position=pos)
    try:
        outedge = int(outedge)
    except (ValueError, TypeError):
        outedge = -1

    if parent is not None:
        parent.add_monosaccharide(base, position=outedge, child_position=min(base.open_attachment_sites()[0]))

    return base, outedge


def parse_linear_code(text, structure_class=Glycan):
    '''
    Parse the character string `text`, extracting GlycoMinds Linear Code-format
    carbohydrate structures, converting them into a |Glycan| object.

    .. note:: Assumes that the structure's root is the right-most residue

    Supports only *concrete* structures.

    The resulting structure will be canonicalized.

    Parameters
    ----------
    text: str
        The string to be parsed

    Returns
    -------
    Glycan

    Raises
    ------
    LinearCodeError:
        When an error is encountered while parsing resulting from malformed syntax
        or structure

    ValueError:
        When a symbol is encountered for which no suitable translation could be found

    KeyError:
        When an unknown symbol is encountered
    '''
    last_outedge = None
    root = None
    last_residue = None
    branch_stack = []
    while len(text) > 0:
        # If starting a new branch
        if text[-1] == ')':
            branch_stack.append((last_residue, root, last_outedge))
            root = None
            last_residue = None
            last_outedge = None
            text = text[:-1]
        # If ending a branch
        elif text[-1] == '(':
            try:
                branch_parent, old_root, old_last_outedge = branch_stack.pop()
                branch_parent.add_monosaccharide(root, position=last_outedge,
                                                 child_position=min(root.open_attachment_sites()[0]))
                root = old_root
                last_residue = branch_parent
                last_outedge = old_last_outedge
                text = text[:-1]
            except IndexError:
                raise LinearCodeError("Bad branching at {}".format(len(text)))
        # Parsing a residue
        else:
            match = re.search(r"([A-Z]+?)(\[[^\]]*?\])?(.)(.)?$", text)
            if match:
                next_residue, outedge = monosaccharide_from_linear_code(text[match.start(): match.end()], last_residue)
                if root is None:
                    last_outedge = outedge
                    root = next_residue
                last_residue = next_residue
                text = text[:match.start()]
            else:
                raise LinearCodeError("Could not identify residue '...{}' at {}".format(text[-10:], len(text)))

    res = structure_class(root=root).reindex()
    res.canonicalize()
    if len(res) > 1:
        return res
    else:
        return res.root


#: Common alias for :func:`to_linear_code`
dumps = to_linear_code

#: Common alias for :func:`parse_linear_code`
loads = parse_linear_code


class LinearCodeParser(ParserInterface):
    def process_result(self, line):
        structure = loads(line)
        return structure


Monosaccharide.register_serializer('linear_code', dumps)
Glycan.register_serializer('linear_code', dumps)



from matplotlib import patches as mpatches
from matplotlib import path as mpath
from matplotlib import transforms as mtransform
import numpy as np

from glypy.utils import groupby


Affine2D = mtransform.Affine2D


class TreeLayoutBase(object):
    '''A base class for defining the layout of a glycan graph on an xy plane

    Attributes
    ----------
    root: :class:`~.DrawTreeNode`
        The root node of the tree to lay out
    '''

    def __init__(self, root, **kwargs):
        self.root = root

    def layout(self, **kwargs):
        """Perform a complete tree layout.

        Parameters
        ----------
        transform: :class:`matplotlib.transforms.Affine2D`
            A transform to apply over the coordinates of each
            node in the tree after the layout algorithm has run.
        **kwargs:
            Forwarded to :meth:`before_layout`, :meth:`layout_tree`,
            and :meth:`after_layout`.

        """
        transform = kwargs.pop("transform", None)
        self.before_layout(**kwargs)
        self.layout_tree(**kwargs)
        self.after_layout(**kwargs)
        if transform is not None:
            self.transform(transform)

    def traverse(self):
        """Traverse the tree recursively.
        """
        return self.root.traverse()

    def layout_tree(self, **kwargs):
        """Assign tree nodes to x-y coordinates.
        """
        raise NotImplementedError()

    def before_layout(self, **kwargs):
        """Prepare the tree before applying layout
        """
        pass

    def after_layout(self, **kwargs):
        """Apply any cleanup after applying layout
        """
        self.root.fix_special_cases()

    def transform(self, transform):
        """Apply an :class:`matplotlib.transforms.Affine2D` transform
        to the x-y coordinates of each node, updating them.

        Parameters
        ----------
        transform : :class:`matplotlib.transforms.Affine2D`
            The transform to apply.
        """
        for node in self.traverse():
            node.x, node.y = transform.transform([node.x, node.y])


def breadth_first_traversal(tree, visited=None):
    if visited is None:
        visited = set()
    if tree.id not in visited:
        visited.add(tree.id)
        yield tree
        for child in tree:
            for descend in breadth_first_traversal(child, visited=visited):
                yield descend


def _build_path_with_radial_bounding_box(node, implicit_radius=0.3):
    pathing = []
    last = [node.x, node.y]
    # build perimeter
    for node in breadth_first_traversal(node):
        pathing.append([node.x, node.y])
        pathing.append(last)
        last = [node.x, node.y]
    codes = [mpath.Path.MOVETO] + [mpath.Path.LINETO for i in pathing[1:]]
    path = mpath.Path(pathing, codes)
    for node in breadth_first_traversal(node):
        center = (node.x, node.y)
        path = path.make_compound_path(path, mpath.Path.circle(center, implicit_radius))
    # for node in breadth_first_traversal(node):
    #     for part in node.patch_node.shapes():
    #         path = path.make_compound_path(path, part.get_path())
    return path


def _build_path_with_centroids(node, **kwargs):
    pathing = []
    last = [node.x, node.y]
    for node in breadth_first_traversal(node):
        pathing.append([node.x, node.y])
        pathing.append(last)
        last = [node.x, node.y]
    codes = [mpath.Path.MOVETO] + [mpath.Path.LINETO for i in pathing[1:]]
    return mpath.Path(pathing, codes)


def _build_path_with_circles(node, implicit_radius=0.3):
    points = []
    for node in breadth_first_traversal(node):
        points.append(mpath.Path.circle((node.x, node.y), implicit_radius))
    return mpath.Path.make_compound_path(*points)


def interpolate(x0, y0, x1, y1, x):
    y = y0 + (x - x0) * ((y1 - y0) / (x1 - x0))
    return y


def make_path(node, implicit_radius=0.3):
    return _build_path_with_circles(node, implicit_radius)


def centroid(path):
    point = np.zeros_like(path.vertices[0])
    c = 0.
    for p in path.vertices:
        point += p
        c += 1
    return point / c


def bounding_box(path, padding=0):
    if not padding:
        padding = 0
    xmin = float('inf')
    xmax = -float('inf')
    ymin = float('inf')
    ymax = -float('inf')
    for p in path.vertices:
        xmin = min(xmin, p[0])
        xmax = max(xmax, p[0])
        ymin = min(ymin, p[1])
        ymax = max(ymax, p[1])
    xmin -= padding
    xmax += padding
    ymin -= padding
    ymax += padding

    return mtransform.Bbox.from_extents(xmin, ymin, xmax, ymax)
    # return xmin, xmax, ymin, ymax


class TransformProxy(object):
    def __init__(self, base_transform, owner_update):
        self.base_transform = base_transform
        self.owner_update = owner_update

    def _copy_transform(self):
        return self.base_transform.frozen()

    def scale(self, *args, **kwargs):
        transform = self._copy_transform()
        transform.scale(*args, **kwargs)
        self.owner_update(transform)
        self.base_transform = transform
        return self

    def rotate(self, *args, **kwargs):
        transform = self._copy_transform()
        transform.rotate(*args, **kwargs)
        self.owner_update(transform)
        self.base_transform = transform
        return self

    def rotate_deg(self, *args, **kwargs):
        transform = self._copy_transform()
        transform.rotate_deg(*args, **kwargs)
        self.owner_update(transform)
        self.base_transform = transform
        return self

    def rotate_around(self, *args, **kwargs):
        transform = self._copy_transform()
        transform.rotate_around(*args, **kwargs)
        self.owner_update(transform)
        self.base_transform = transform
        return self

    def rotate_deg_around(self, *args, **kwargs):
        transform = self._copy_transform()
        transform.rotate_deg_around(*args, **kwargs)
        self.owner_update(transform)
        self.base_transform = transform
        return self

    def __getattr__(self, name):
        if name == 'base_transform':
            raise AttributeError("Accessing Base Transform Too Early")
        return getattr(self.base_transform, name)


def l2_distance(a, b):
    return sum([(ai - bi) ** 2 for ai, bi in zip(a, b)])


def l1_distance(a, b):
    return sum([abs(ai - bi) for ai, bi in zip(a, b)])


# def shift_path_2d(path, hits):
#     centers = map(centroid, hits)
#     stacked = np.vstack(centers)
#     min_x = stacked[:, 0].min()
#     max_x = stacked[:, 0].max()
#     min_y = stacked[:, 1].min()
#     max_y = stacked[:, 1].max()
#     new_center = np.array([(min_x + max_x)/2., (min_y + max_y)/2.]) * 0.2
#     current_center = centroid(path)
#     distance = current_center - new_center
#     path = path.transformed(Affine2D().translate(*distance))
#     return path

# def shift_path_away(path, hit):
#     center = centroid(path)
#     distance = ((centroid(hit) - center) * 0.2)
#     return path.transformed(Affine2D().translate(*distance))



# def find_overlaps(path, objs):
#     hits = []
#     for obj in objs:
#         if path.intersects_path(obj):
#             hits.append(obj)
#     return hits

# def minimize_overlap(path, objs, max_distance=1.0):
#     hits = find_overlaps(path, objs)
#     # We hit two or more objects
#     if len(hits) > 1:
#         path = shift_path_2d(path, hits)
#     elif len(hits) == 1:
#         path = shift_path_away(path, hits[0])
#     again_hits = find_overlaps(path, objs)
#     return path
